from django.db.models.signals import post_save
from django.utils.translation import gettext_lazy as _
from swapper import get_model_name, load_model

from openwisp_utils.admin_theme.menu import register_menu_group
from openwisp_utils.api.apps import ApiAppConfig
from openwisp_utils.utils import default_or_test

from . import settings as app_settings


class FirmwareUpdaterConfig(ApiAppConfig):
    name = 'openwisp_firmware_upgrader'
    label = 'firmware_upgrader'
    verbose_name = _('Firmware Management')
    default_auto_field = 'django.db.models.AutoField'

    API_ENABLED = app_settings.FIRMWARE_UPGRADER_API
    REST_FRAMEWORK_SETTINGS = {
        'DEFAULT_THROTTLE_RATES': {
            'firmware_upgrader': default_or_test('1000/minute', None)
        },
    }

    def ready(self, *args, **kwargs):
        super().ready(*args, **kwargs)
        self.register_menu_groups()
        self.connect_device_signals()

    def register_menu_groups(self):
        register_menu_group(
            position=100,
            config={
                'label': _('Firmware'),
                'items': {
                    1: {
                        'label': _('Builds'),
                        'model': get_model_name(self.label, 'Build'),
                        'name': 'changelist',
                        'icon': 'ow-build',
                    },
                    2: {
                        'label': _('Categories'),
                        'model': get_model_name(self.label, 'Category'),
                        'name': 'changelist',
                        'icon': 'ow-category',
                    },
                    3: {
                        'label': _('Mass Upgrade Operations'),
                        'model': get_model_name(self.label, 'BatchUpgradeOperation'),
                        'name': 'changelist',
                        'icon': 'ow-mass-upgrade',
                    },
                },
                'icon': 'ow-firmware',
            },
        )

    def connect_device_signals(self):
        DeviceConnection = load_model('connection', 'DeviceConnection')
        DeviceFirmware = load_model('firmware_upgrader', 'DeviceFirmware')
        FirmwareImage = load_model('firmware_upgrader', 'FirmwareImage')
        post_save.connect(
            DeviceFirmware.auto_add_device_firmware_to_device,
            sender=DeviceConnection,
            dispatch_uid='connection.auto_add_device_firmware',
        )
        post_save.connect(
            DeviceFirmware.auto_create_device_firmwares,
            sender=FirmwareImage,
            dispatch_uid='firmware_image.auto_add_device_firmwares',
        )


del ApiAppConfig



from private_storage.views import PrivateStorageDetailView

from ..swapper import load_model


class FirmwareImageDownloadView(PrivateStorageDetailView):
    model = load_model('FirmwareImage')
    model_file_field = 'file'

    slug_field = 'file'
    slug_url_kwarg = 'path'

    def can_access_file(self, private_file):
        user = private_file.request.user
        return user.is_superuser or (
            user.is_staff and user.is_manager(self.object.build.category.organization)
        )


firmware_image_download = FirmwareImageDownloadView.as_view()



# Generated by Django 3.0.8 on 2020-07-07 17:00

from django.db import migrations

from . import create_device_firmware_for_connections


def create_device_firmware_for_connections_helper(apps, schema_editor):
    app_label = 'firmware_upgrader'
    create_device_firmware_for_connections(apps, schema_editor, app_label)


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0005_build_os_identifier'),
    ]

    operations = [
        migrations.RunPython(
            create_device_firmware_for_connections_helper,
            reverse_code=migrations.RunPython.noop,
        ),
    ]



# Generated by Django 3.1.7 on 2021-03-27 14:39

from urllib.parse import urljoin

import private_storage.fields
import private_storage.storage.files
from django.db import migrations

import openwisp_firmware_upgrader.base.models

from ..settings import FIRMWARE_API_BASEURL, IMAGE_URL_PATH


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0007_nullable_upgrade_operation_image'),
    ]

    operations = [
        migrations.AlterField(
            model_name='firmwareimage',
            name='file',
            field=private_storage.fields.PrivateFileField(
                storage=private_storage.storage.files.PrivateFileSystemStorage(
                    base_url=urljoin(FIRMWARE_API_BASEURL, IMAGE_URL_PATH),
                ),
                upload_to=openwisp_firmware_upgrader.base.models.get_build_directory,
                verbose_name='File',
            ),
        ),
    ]



from unittest.mock import patch

from django.test import TestCase

from .. import settings as app_settings
from ..utils import get_upgrader_class_from_device_connection
from .base import TestUpgraderMixin


class TestUtils(TestUpgraderMixin, TestCase):
    @patch('logging.Logger.exception')
    def test_get_upgrader_class_from_device_connection(self, mocked_logger):
        device_conn = self._create_device_connection()

        with self.subTest('Test upgrader is not configured in "UPGRADERS_MAP"'):
            with patch.object(app_settings, 'UPGRADERS_MAP', {}):
                upgrader_class = get_upgrader_class_from_device_connection(device_conn)
                self.assertEqual(upgrader_class, None)
                mocked_logger.assert_called()

        mocked_logger.reset_mock()

        with self.subTest('Test upgrader is not configured in "UPGRADERS_MAP"'):
            with patch.object(
                app_settings,
                'UPGRADERS_MAP',
                {
                    device_conn.update_strategy: 'openwisp_firmware_upgrader.upgraders.invalid'
                },
            ):
                upgrader_class = get_upgrader_class_from_device_connection(device_conn)
                self.assertEqual(upgrader_class, None)
                mocked_logger.assert_called()



from django.contrib.auth.management import create_permissions
from django.contrib.auth.models import Permission
from swapper import load_model

DeviceConnection = load_model('connection', 'DeviceConnection')
DeviceFirmware = load_model('firmware_upgrader', 'DeviceFirmware')


def create_default_permissions(apps, schema_editor):
    for app_config in apps.get_app_configs():
        app_config.models_module = True
        create_permissions(app_config, apps=apps, verbosity=0)
        app_config.models_module = None


def create_permissions_for_default_groups(apps, schema_editor, app_label):
    create_default_permissions(apps, schema_editor)
    Group = apps.get_model('openwisp_users', 'Group')

    try:
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')
    # consider failures custom cases
    # that do not have to be dealt with
    except Group.DoesNotExist:
        return

    operators_read_only_admins_manage = [
        'build',
        'devicefirmware',
        'firmwareimage',
        'batchupgradeoperation',
        'upgradeoperation',
    ]
    admins_can_manage = ['category']
    manage_operations = ['add', 'change', 'delete']

    for action in manage_operations:
        for model_name in admins_can_manage:
            permission = Permission.objects.get(
                codename='{}_{}'.format(action, model_name),
                content_type__app_label=app_label,
            )
            admin.permissions.add(permission.pk)
    for model_name in operators_read_only_admins_manage:
        try:
            permission = Permission.objects.get(
                codename='view_{}'.format(model_name), content_type__app_label=app_label
            )
            operator.permissions.add(permission.pk)
        except Permission.DoesNotExist:
            pass
        for action in manage_operations:
            permission = Permission.objects.get(
                codename='{}_{}'.format(action, model_name),
                content_type__app_label=app_label,
            )
            admin.permissions.add(permission.pk)


def create_device_firmware_for_connections(apps, schema_editor, app_label):
    for device_connection in DeviceConnection.objects.all():
        DeviceFirmware.create_for_device(device_connection.device)



from urllib.parse import urljoin

from private_storage.storage.files import PrivateFileSystemStorage

from ..settings import FIRMWARE_API_BASEURL, IMAGE_URL_PATH

file_system_private_storage = PrivateFileSystemStorage(
    base_url=urljoin(FIRMWARE_API_BASEURL, IMAGE_URL_PATH)
)



import swapper
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _
from rest_framework import serializers

from openwisp_users.api.mixins import FilterSerializerByOrgManaged
from openwisp_utils.api.serializers import ValidatedModelSerializer

from ..swapper import load_model

BatchUpgradeOperation = load_model('BatchUpgradeOperation')
Build = load_model('Build')
Category = load_model('Category')
FirmwareImage = load_model('FirmwareImage')
UpgradeOperation = load_model('UpgradeOperation')
DeviceFirmware = load_model('DeviceFirmware')
Device = swapper.load_model('config', 'Device')


class BaseMeta:
    read_only_fields = ['created', 'modified']


class BaseSerializer(FilterSerializerByOrgManaged, ValidatedModelSerializer):
    pass


class CategorySerializer(BaseSerializer):
    class Meta(BaseMeta):
        model = Category
        fields = '__all__'


class CategoryRelationSerializer(BaseSerializer):
    class Meta:
        model = Category
        fields = ['name', 'organization']


class FirmwareImageSerializer(BaseSerializer):
    def validate(self, data):
        data['build'] = self.context['view'].get_parent_queryset().get()
        return super().validate(data)

    class Meta(BaseMeta):
        model = FirmwareImage
        fields = '__all__'
        read_only_fields = BaseMeta.read_only_fields + ['build']


class BuildSerializer(BaseSerializer):
    category_relation = CategoryRelationSerializer(read_only=True, source='category')

    class Meta(BaseMeta):
        model = Build
        fields = '__all__'


class UpgradeOperationSerializer(serializers.ModelSerializer):
    class Meta:
        model = UpgradeOperation
        fields = ('id', 'device', 'image', 'status', 'log', 'modified', 'created')


class DeviceUpgradeOperationSerializer(serializers.ModelSerializer):
    class Meta:
        model = UpgradeOperation
        fields = ('id', 'device', 'image', 'status', 'log', 'modified')


class BatchUpgradeOperationListSerializer(BaseSerializer):
    build = BuildSerializer(read_only=True)

    class Meta:
        model = BatchUpgradeOperation
        fields = '__all__'


class BatchUpgradeOperationSerializer(BatchUpgradeOperationListSerializer):
    progress_report = serializers.CharField(max_length=200)
    success_rate = serializers.IntegerField(read_only=True)
    failed_rate = serializers.IntegerField(read_only=True)
    aborted_rate = serializers.IntegerField(read_only=True)
    upgradeoperations = UpgradeOperationSerializer(
        read_only=True, source='upgradeoperation_set', many=True
    )

    class Meta:
        model = BatchUpgradeOperation
        fields = '__all__'


class DeviceFirmwareSerializer(ValidatedModelSerializer):
    class Meta:
        model = DeviceFirmware
        fields = ('id', 'image', 'installed', 'modified')
        read_only_fields = ('installed', 'modified')

    def validate(self, data):
        if not data.get('device'):
            device_id = self.context.get('device_id')
            device = self._get_device_object(device_id)
            data.update({'device': device})
        image = data.get('image')
        device = data.get('device')
        if (
            image
            and device
            and image.build.category.organization != device.organization
        ):
            raise ValidationError(
                {
                    'image': _(
                        'The organization of the image doesn\'t '
                        'match the organization of the device'
                    )
                }
            )
        return super().validate(data)

    def _get_device_object(self, device_id):
        try:
            device = Device.objects.get(id=device_id)
            return device
        except Device.DoesNotExist:
            return None



from urllib.parse import urljoin

from django.urls import path

from ..settings import IMAGE_URL_PATH
from . import views

urlpatterns = [
    path(
        # Use "path" URL kwarg to make it consistent with
        # django-private-storage. Otherwise, the S3 reverse
        # proxy feature of django-private-storage does
        # not work.
        urljoin(IMAGE_URL_PATH, '<path:path>'),
        views.firmware_image_download,
        name='serve_private_file',
    ),
]



# Generated by Django 3.2.18 on 2023-03-03 21:37

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0008_firmware_image_baseurl'),
    ]

    operations = [
        migrations.AddField(
            model_name='batchupgradeoperation',
            name='upgrade_options',
            field=models.JSONField(default=dict, blank=True),
        ),
        migrations.AddField(
            model_name='upgradeoperation',
            name='upgrade_options',
            field=models.JSONField(default=dict, blank=True),
        ),
    ]



# Generated by Django 3.1.5 on 2021-01-21 16:57

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0006_create_device_firmware'),
    ]

    operations = [
        migrations.AlterField(
            model_name='upgradeoperation',
            name='image',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                to=settings.FIRMWARE_UPGRADER_FIRMWAREIMAGE_MODEL,
            ),
        ),
    ]



from django.db import migrations

from . import create_permissions_for_default_groups


def create_permissions_for_default_groups_helper(apps, schema_editor):
    app_label = 'firmware_upgrader'
    create_permissions_for_default_groups(apps, schema_editor, app_label)


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(
            create_permissions_for_default_groups_helper,
            reverse_code=migrations.RunPython.noop,
        )
    ]



from unittest.mock import patch

import swapper
from django.contrib.staticfiles.testing import StaticLiveServerTestCase
from django.core.management import call_command
from django.urls.base import reverse
from reversion.models import Version
from selenium.common.exceptions import TimeoutException, UnexpectedAlertPresentException
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import Select, WebDriverWait

from openwisp_firmware_upgrader.hardware import REVERSE_FIRMWARE_IMAGE_MAP
from openwisp_firmware_upgrader.tests.base import TestUpgraderMixin
from openwisp_utils.test_selenium_mixins import SeleniumTestMixin
from openwisp_utils.tests import capture_any_output

from ..swapper import load_model
from ..upgraders.openwisp import OpenWrt

Device = swapper.load_model('config', 'Device')
DeviceConnection = swapper.load_model('connection', 'DeviceConnection')
UpgradeOperation = load_model('UpgradeOperation')
DeviceFirmware = load_model('DeviceFirmware')
BatchUpgradeOperation = load_model('BatchUpgradeOperation')


class TestDeviceAdmin(TestUpgraderMixin, SeleniumTestMixin, StaticLiveServerTestCase):
    config_app_label = 'config'
    firmware_app_label = 'firmware_upgrader'
    os = 'OpenWrt 19.07-SNAPSHOT r11061-6ffd4d8a4d'
    image_type = REVERSE_FIRMWARE_IMAGE_MAP['YunCore XD3200']

    def _set_up_env(self):
        org = self._get_org()
        category = self._get_category(organization=org)
        build1 = self._create_build(category=category, version='0.1', os=self.os)
        build2 = self._create_build(
            category=category, version='0.2', os='OpenWrt 21.03'
        )
        image1 = self._create_firmware_image(build=build1, type=self.image_type)
        image2 = self._create_firmware_image(build=build2, type=self.image_type)
        self._create_credentials(auto_add=True, organization=org)
        device = self._create_device(
            os=self.os, model=image2.boards[0], organization=org
        )
        self._create_config(device=device)
        return org, category, build1, build2, image1, image2, device

    def setUp(self):
        self.admin = self._create_admin(
            username=self.admin_username, password=self.admin_password
        )

    def tearDown(self):
        # Accept unsaved changes alert to allow other tests to run
        try:
            self.web_driver.refresh()
        except UnexpectedAlertPresentException:
            self.web_driver.switch_to_alert().accept()
        else:
            try:
                WebDriverWait(self.web_driver, 1).until(EC.alert_is_present())
            except TimeoutException:
                pass
            else:
                self.web_driver.switch_to_alert().accept()
        self.web_driver.refresh()
        WebDriverWait(self.web_driver, 2).until(
            EC.visibility_of_element_located((By.XPATH, '//*[@id="site-name"]'))
        )

    @capture_any_output()
    def test_restoring_deleted_device(self):
        org = self._get_org()
        category = self._get_category(organization=org)
        build = self._create_build(category=category, version='0.1', os=self.os)
        image = self._create_firmware_image(build=build, type=self.image_type)
        self._create_credentials(auto_add=True, organization=org)
        device = self._create_device(
            os=self.os, model=image.boards[0], organization=org
        )
        config = self._create_config(device=device)
        self.assertEqual(Device.objects.count(), 1)
        self.assertEqual(DeviceConnection.objects.count(), 1)
        self.assertEqual(DeviceFirmware.objects.count(), 1)

        call_command('createinitialrevisions')

        self.login()
        device.deactivate()
        config.set_status_deactivated()
        # Delete the device
        self.open(
            reverse(f'admin:{self.config_app_label}_device_delete', args=[device.id])
        )
        self.web_driver.find_element(
            by=By.CSS_SELECTOR, value='#content form input[type="submit"]'
        ).click()
        self.assertEqual(Device.objects.count(), 0)
        self.assertEqual(DeviceConnection.objects.count(), 0)
        self.assertEqual(DeviceFirmware.objects.count(), 0)

        version_obj = Version.objects.get_deleted(model=Device).first()

        # Restore deleted device
        self.open(
            reverse(
                f'admin:{self.config_app_label}_device_recover', args=[version_obj.id]
            )
        )
        self.web_driver.find_element(
            by=By.XPATH, value='//*[@id="device_form"]/div/div[1]/input[1]'
        ).click()
        try:
            WebDriverWait(self.web_driver, 5).until(
                EC.url_to_be(f'{self.live_server_url}/admin/config/device/')
            )
        except TimeoutException:
            self.fail('Deleted device was not restored')

        self.assertEqual(Device.objects.count(), 1)
        self.assertEqual(DeviceConnection.objects.count(), 1)
        self.assertEqual(DeviceFirmware.objects.count(), 1)

    @capture_any_output()
    @patch(
        'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt.upgrade',
        return_value=True,
    )
    @patch(
        'openwisp_controller.connection.models.DeviceConnection.connect',
        return_value=True,
    )
    def test_device_firmware_upgrade_options(self, *args):
        def save_device():
            self.web_driver.find_element(
                by=By.XPATH, value='//*[@id="device_form"]/div/div[1]/input[3]'
            ).click()

        _, _, _, _, _, image, device = self._set_up_env()
        self.login()
        self.open(
            '{}#devicefirmware-group'.format(
                reverse(
                    f'admin:{self.config_app_label}_device_change', args=[device.id]
                )
            )
        )
        # JSONSchema Editor should not be rendered without a change in the image field
        WebDriverWait(self.web_driver, 1).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, '#devicefirmware-group .jsoneditor-wrapper')
            )
        )
        image_select = Select(
            self.web_driver.find_element(by=By.ID, value='id_devicefirmware-0-image')
        )
        image_select.select_by_value(str(image.pk))
        # JSONSchema configuration editor should not be rendered
        WebDriverWait(self.web_driver, 1).until(
            EC.invisibility_of_element_located(
                (
                    By.XPATH,
                    '//*[@id="id_devicefirmware-0-upgrade_options_jsoneditor"]/div/h3/span[4]/input',
                )
            )
        )
        # Select "None" image should hide JSONSchema Editor
        image_select.select_by_value('')
        WebDriverWait(self.web_driver, 1).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, '#id_devicefirmware-0-upgrade_options_jsoneditor')
            )
        )

        # Select "build2" image
        image_select.select_by_value(str(image.pk))
        # Enable '-c' option
        self.web_driver.find_element(
            by=By.XPATH,
            value='//*[@id="id_devicefirmware-0-upgrade_options_jsoneditor"]'
            '/div/div[2]/div/div/div[1]/div/div[1]/label/input',
        ).click()
        # Enable '-F' option
        self.web_driver.find_element(
            by=By.XPATH,
            value='//*[@id="id_devicefirmware-0-upgrade_options_jsoneditor"]'
            '/div/div[2]/div/div/div[7]/div/div[1]/label/input',
        ).click()
        save_device()

        # Delete DeviceFirmware
        self.web_driver.find_element(
            by=By.ID, value='id_devicefirmware-0-DELETE'
        ).click()
        save_device()

        # When adding firmware to the device for the first time,
        # JSONSchema editor should be rendered only when the image
        # is selected
        self.web_driver.find_element(
            by=By.XPATH, value='//*[@id="devicefirmware-group"]/fieldset/div[2]/a'
        ).click()
        # JSONSchema Editor should not be rendered without a change in the image field
        WebDriverWait(self.web_driver, 1).until(
            EC.invisibility_of_element_located(
                (By.CSS_SELECTOR, '#devicefirmware-group .jsoneditor-wrapper')
            )
        )
        image_select = Select(
            self.web_driver.find_element(by=By.ID, value='id_devicefirmware-0-image')
        )
        image_select.select_by_index(1)
        try:
            WebDriverWait(self.web_driver, 1).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, '#devicefirmware-group .jsoneditor-wrapper')
                )
            )
        except TimeoutError:
            self.fail('JSONSchema editor not shown after changing firmware image')
        save_device()

    @capture_any_output()
    @patch(
        'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt.upgrade',
        return_value=True,
    )
    @patch(
        'openwisp_controller.connection.models.DeviceConnection.connect',
        return_value=True,
    )
    def test_batch_upgrade_upgrade_options(self, *args):
        _, _, _, build2, _, _, _ = self._set_up_env()
        self.login()
        self.open(
            reverse(f'admin:{self.firmware_app_label}_build_change', args=[build2.id])
        )
        # Launch mass upgrade operation
        self.web_driver.find_element(
            by=By.CSS_SELECTOR,
            value='.title-wrapper .object-tools form button[type="submit"]',
        ).click()

        # Ensure JSONSchema form is rendered
        try:
            WebDriverWait(self.web_driver, 1).until(
                EC.visibility_of_element_located(
                    (By.CSS_SELECTOR, '.jsoneditor-wrapper')
                )
            )
        except TimeoutError:
            self.fail('JSONSchema editor not shown after changing firmware image')
        # JSONSchema configuration editor should not be rendered
        WebDriverWait(self.web_driver, 1).until(
            EC.invisibility_of_element_located(
                (
                    By.XPATH,
                    '//*[@id="id_devicefirmware-0-upgrade_options_jsoneditor"]/div/h3/span[4]/input',
                )
            )
        )
        # Disable -c flag
        self.web_driver.find_element(
            by=By.XPATH,
            value='//*[@id="id_upgrade_options_jsoneditor"]/div/div[2]/div/div/div[1]/div/div[1]/label/input',
        ).click()
        # Enable -n flag
        self.web_driver.find_element(
            by=By.XPATH,
            value='//*[@id="id_upgrade_options_jsoneditor"]/div/div[2]/div/div/div[3]/div/div[1]/label/input',
        ).click()
        # Upgrade all devices
        self.web_driver.find_element(
            by=By.CSS_SELECTOR, value='input[name="upgrade_all"]'
        ).click()
        self.assertEqual(
            BatchUpgradeOperation.objects.filter(
                upgrade_options={
                    'c': False,
                    'o': False,
                    'n': True,
                    'u': False,
                    'p': False,
                    'k': False,
                    'F': False,
                }
            ).count(),
            1,
        )
        self.assertEqual(
            UpgradeOperation.objects.filter(
                upgrade_options={
                    'c': False,
                    'o': False,
                    'n': True,
                    'u': False,
                    'p': False,
                    'k': False,
                    'F': False,
                }
            ).count(),
            1,
        )

    @capture_any_output()
    @patch(
        'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt.upgrade',
        return_value=True,
    )
    @patch(
        'openwisp_controller.connection.models.DeviceConnection.connect',
        return_value=True,
    )
    @patch.object(OpenWrt, 'SCHEMA', None)
    def test_upgrader_with_unsupported_upgrade_options(self, *args):
        org, category, build1, build2, image1, image2, device = self._set_up_env()
        self.login()

        with self.subTest('Test DeviceFirmware'):
            self.open(
                '{}#devicefirmware-group'.format(
                    reverse(
                        f'admin:{self.config_app_label}_device_change', args=[device.id]
                    )
                )
            )
            image_select = Select(
                self.web_driver.find_element(
                    by=By.ID, value='id_devicefirmware-0-image'
                )
            )
            image_select.select_by_value(str(image2.pk))
            # Ensure JSONSchema editor is not rendered because
            # the upgrader does not define a schema
            WebDriverWait(self.web_driver, 1).until(
                EC.invisibility_of_element_located(
                    (By.CSS_SELECTOR, '#devicefirmware-group .jsoneditor-wrapper')
                )
            )
            self.web_driver.find_element(
                by=By.XPATH, value='//*[@id="device_form"]/div/div[1]/input[3]'
            ).click()
            self.assertEqual(
                UpgradeOperation.objects.filter(upgrade_options={}).count(), 1
            )
        DeviceFirmware.objects.all().delete()
        UpgradeOperation.objects.all().delete()

        with self.subTest('Test BatchUpgradeOperation'):
            self.open(
                reverse(
                    f'admin:{self.firmware_app_label}_build_change', args=[build2.id]
                )
            )
            # Launch mass upgrade operation
            self.web_driver.find_element(
                by=By.CSS_SELECTOR,
                value='.title-wrapper .object-tools form button[type="submit"]',
            ).click()
            # Ensure JSONSchema editor is not rendered because
            # the upgrader does not define a schema
            WebDriverWait(self.web_driver, 1).until(
                EC.invisibility_of_element_located(
                    (By.CSS_SELECTOR, '#devicefirmware-group .jsoneditor-wrapper')
                )
            )
            # Upgrade all devices
            self.web_driver.find_element(
                by=By.CSS_SELECTOR, value='input[name="upgrade_all"]'
            ).click()
            self.assertEqual(
                UpgradeOperation.objects.filter(upgrade_options={}).count(), 1
            )



from django.utils.translation import gettext_lazy as _
from django_filters import rest_framework as filters

from openwisp_users.api.mixins import FilterDjangoByOrgManaged

from ..swapper import load_model

UpgradeOperation = load_model('UpgradeOperation')


class UpgradeOperationFilter(FilterDjangoByOrgManaged):
    device = filters.CharFilter(
        field_name='device',
    )
    image = filters.CharFilter(
        field_name='image',
    )

    def _set_valid_filterform_lables(self):
        self.filters['device__organization'].label = _('Organization')
        self.filters['device__organization__slug'].label = _('Organization slug')

    def __init__(self, *args, **kwargs):
        super(UpgradeOperationFilter, self).__init__(*args, **kwargs)
        self._set_valid_filterform_lables()

    class Meta:
        model = UpgradeOperation
        fields = [
            'device__organization',
            'device__organization__slug',
            'device',
            'image',
            'status',
        ]


class DeviceUpgradeOperationFilter(FilterDjangoByOrgManaged):
    class Meta:
        model = UpgradeOperation
        fields = ['status']



from django import forms

from openwisp_controller.config.widgets import JsonSchemaWidget as BaseJsonSchemaWidget

from .swapper import load_model

UpgradeOperation = load_model('UpgradeOperation')
DeviceFirmware = load_model('DeviceFirmware')


class FirmwareSchemaWidget(BaseJsonSchemaWidget):
    schema_view_name = None
    app_label_model = (
        f'{UpgradeOperation._meta.app_label}_{UpgradeOperation._meta.model_name}'
    )
    netjsonconfig_hint = False
    advanced_mode = False
    extra_attrs = {
        'data-show-errors': 'never',
        'class': 'manual',
    }

    @property
    def media(self):
        return super().media + forms.Media(
            css={'all': ['firmware-upgrader/css/upgrade-options.css']}
        )



class FirmwareUpgraderException(Exception):
    pass


class FirmwareUpgradeOptionsException(Exception):
    pass


class UpgradeNotNeeded(FirmwareUpgraderException):
    """
    Raised when the upgrade is not needed
    """


class UpgradeAborted(FirmwareUpgraderException):
    """
    Raised when the upgrade has to be flagged as aborted
    """


class ReconnectionFailed(FirmwareUpgraderException):
    """
    Raised when the reconnection after the upgrade fails
    """


class RecoverableFailure(FirmwareUpgraderException):
    """
    Raised when the upgrade can be retried
    """



import json
from datetime import timedelta
from unittest import mock

import swapper
from django.contrib.admin.helpers import ACTION_CHECKBOX_NAME
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Permission
from django.test import RequestFactory, TestCase, TransactionTestCase
from django.urls import reverse
from django.utils.timezone import localtime

from openwisp_controller.config.tests.test_admin import TestAdmin as TestConfigAdmin
from openwisp_controller.connection import settings as conn_settings
from openwisp_firmware_upgrader.admin import (
    BuildAdmin,
    DeviceAdmin,
    DeviceFirmwareForm,
    DeviceFirmwareInline,
    DeviceUpgradeOperationInline,
    FirmwareImageInline,
    admin,
)
from openwisp_users.tests.utils import TestMultitenantAdminMixin
from openwisp_utils.tests import AdminActionPermTestMixin, capture_stderr

from ..hardware import REVERSE_FIRMWARE_IMAGE_MAP
from ..swapper import load_model
from ..upgraders.openwisp import OpenWisp1
from .base import TestUpgraderMixin

User = get_user_model()

Build = load_model('Build')
Category = load_model('Category')
DeviceFirmware = load_model('DeviceFirmware')
FirmwareImage = load_model('FirmwareImage')
UpgradeOperation = load_model('UpgradeOperation')
BatchUpgradeOperation = load_model('BatchUpgradeOperation')
Device = swapper.load_model('config', 'Device')


class MockRequest:
    pass


class BaseTestAdmin(TestMultitenantAdminMixin, TestUpgraderMixin):
    app_label = 'firmware_upgrader'
    _device_params = TestConfigAdmin._device_params.copy()
    _device_params.update(
        {
            'devicefirmware-0-image': '',
            'devicefirmware-0-id': '',
            'devicefirmware-TOTAL_FORMS': 0,
            'devicefirmware-INITIAL_FORMS': 0,
            'devicefirmware-MIN_NUM_FORMS': 0,
            'devicefirmware-MAX_NUM_FORMS': 1,
            'deviceconnection_set-TOTAL_FORMS': 1,
            'deviceconnection_set-INITIAL_FORMS': 1,
            'devicelocation-TOTAL_FORMS': 1,
            'devicelocation-INITIAL_FORMS': 0,
            'devicelocation-MIN_NUM_FORMS': 0,
            'devicelocation-MAX_NUM_FORMS': 1,
            'config-INITIAL_FORMS': 1,
        }
    )

    def setUp(self, *args, **kwargs):
        super().setUp(*args, **kwargs)
        self.factory = RequestFactory()

    def make_device_admin_request(self, pk):
        return self.factory.get(reverse('admin:config_device_change', args=[pk]))

    @property
    def build_list_url(self):
        return reverse(f'admin:{self.app_label}_build_changelist')


class TestAdmin(BaseTestAdmin, TestCase):
    def test_build_list(self):
        self._login()
        build = self._create_build()
        r = self.client.get(self.build_list_url)
        self.assertContains(r, str(build))

    def test_build_list_upgrade_action(self):
        self._login()
        self._create_build()
        r = self.client.get(self.build_list_url)
        self.assertContains(r, '<option value="upgrade_selected">')

    def test_upgrade_build_admin(self):
        self._login()
        b = self._create_build()
        path = reverse(f'admin:{self.app_label}_build_change', args=[b.pk])
        r = self.client.get(path)
        self.assertEqual(r.status_code, 200)
        self.assertContains(r, 'Launch mass upgrade operation')

    def test_upgrade_selected_error(self):
        self._login()
        b1 = self._create_build()
        b2 = self._create_build(version='0.2', category=b1.category)
        r = self.client.post(
            self.build_list_url,
            {'action': 'upgrade_selected', ACTION_CHECKBOX_NAME: (b1.pk, b2.pk)},
            follow=True,
        )
        self.assertContains(r, '<li class="error">')
        self.assertContains(
            r, 'only a single mass upgrade operation at time is supported'
        )

    def test_upgrade_intermediate_page_related(self):
        self._login()
        env = self._create_upgrade_env()
        with self.assertNumQueries(13):
            r = self.client.post(
                self.build_list_url,
                {
                    'action': 'upgrade_selected',
                    ACTION_CHECKBOX_NAME: (env['build2'].pk,),
                },
                follow=True,
            )
        self.assertContains(r, 'Devices related to build')
        self.assertNotContains(r, 'has never upgraded yet')
        self.assertNotContains(r, '<input type="submit" name="upgrade_related"')

    def test_upgrade_intermediate_page_firmwareless(self):
        self._login()
        env = self._create_upgrade_env(device_firmware=False)
        with self.assertNumQueries(12):
            r = self.client.post(
                self.build_list_url,
                {
                    'action': 'upgrade_selected',
                    ACTION_CHECKBOX_NAME: (env['build2'].pk,),
                },
                follow=True,
            )
        self.assertNotContains(r, 'Devices related to build')
        self.assertContains(r, 'has never upgraded yet')
        self.assertNotContains(r, '<input type="submit" name="upgrade_related"')
        self.assertContains(r, '<input type="submit" name="upgrade_all"')

    def test_view_device_administrator(self):
        device_fw = self._create_device_firmware()
        org = self._get_org()
        self._create_administrator(organizations=[org])
        self._login(username='administrator', password='tester')
        url = reverse('admin:config_device_change', args=[device_fw.device_id])
        r = self.client.get(url)
        self.assertContains(r, str(device_fw.image_id))

    def test_firmware_image_has_change_permission(self):
        request = MockRequest()
        request.user = User.objects.first()
        env = self._create_upgrade_env()
        self.assertIn(FirmwareImageInline, BuildAdmin.inlines)
        inline = FirmwareImageInline(FirmwareImage, admin.site)
        self.assertIsInstance(inline, FirmwareImageInline)
        self.assertIs(inline.has_change_permission(request), True)
        self.assertIs(inline.has_change_permission(request, obj=env['image1a']), False)

    def test_device_firmware_inline_has_add_permission(self):
        device_fw = self._create_device_firmware()
        device = device_fw.device
        request = self.make_device_admin_request(device.pk)
        request.user = User.objects.first()
        inline = DeviceFirmwareInline(Device, admin.site)
        self.assertTrue(inline.has_add_permission(request, obj=None))
        self.assertTrue(inline.has_add_permission(request, obj=device))
        self.assertIsInstance(inline, DeviceFirmwareInline)
        deviceadmin = DeviceAdmin(model=Device, admin_site=admin.site)
        self.assertIn(
            DeviceFirmwareInline, deviceadmin.get_inlines(request, obj=device)
        )

    def test_device_firmware_inline(self):
        device_fw = self._create_device_firmware()
        device = device_fw.device
        request = self.make_device_admin_request(device.pk)
        request.user = User.objects.first()
        deviceadmin = DeviceAdmin(model=Device, admin_site=admin.site)
        self.assertNotIn(
            DeviceFirmwareInline, deviceadmin.get_inlines(request, obj=None)
        )
        self.assertIn(
            DeviceFirmwareInline, deviceadmin.get_inlines(request, obj=device)
        )

    def _prepare_image_qs_test_env(self):
        device_fw = self._create_device_firmware()
        device = device_fw.device
        request = self.make_device_admin_request(device.pk)
        request.user = User.objects.first()
        org2 = self._create_org(name='org2', slug='org2')
        category_org2 = self._create_category(organization=org2, name='org2')
        build_org2 = self._create_build(category=category_org2)
        img_org2 = self._create_firmware_image(build=build_org2)
        yuncore = self._create_firmware_image(
            build=device_fw.image.build,
            type=REVERSE_FIRMWARE_IMAGE_MAP['YunCore XD3200'],
        )
        mesh_category = self._create_category(
            name='mesh', organization=device.organization
        )
        mesh_build = self._create_build(category=mesh_category)
        mesh_image = self._create_firmware_image(build=mesh_build)
        return device, device_fw, img_org2, yuncore, mesh_image

    def test_image_queryset_existing_device_firmware(self):
        (
            device,
            device_fw,
            img_org2,
            yuncore,
            mesh_image,
        ) = self._prepare_image_qs_test_env()
        # existing DeviceFirmware
        # restricts images to category of image in used
        form = DeviceFirmwareForm(device=device, instance=device_fw)
        self.assertEqual(form.fields['image'].queryset.count(), 1)
        self.assertIn(device_fw.image, form.fields['image'].queryset)
        self.assertNotIn(img_org2, form.fields['image'].queryset)

    def test_image_queryset_new_device_firmware(self):
        (
            device,
            device_fw,
            img_org2,
            yuncore,
            mesh_image,
        ) = self._prepare_image_qs_test_env()
        # new DeviceFirmware
        # shows all the categories related to the model
        form = DeviceFirmwareForm(device=device)
        self.assertEqual(form.fields['image'].queryset.count(), 2)
        self.assertIn(device_fw.image, form.fields['image'].queryset)
        self.assertIn(mesh_image, form.fields['image'].queryset)
        self.assertNotIn(img_org2, form.fields['image'].queryset)

    def test_image_queryset_no_model(self):
        (
            device,
            device_fw,
            img_org2,
            yuncore,
            mesh_image,
        ) = self._prepare_image_qs_test_env()
        # existing DeviceFirmware
        # if no model specified, get all models available
        device.model = ''
        device.save()
        form = DeviceFirmwareForm(device=device, instance=device_fw)
        self.assertEqual(form.fields['image'].queryset.count(), 2)
        self.assertIn(yuncore, form.fields['image'].queryset)
        self.assertNotIn(img_org2, form.fields['image'].queryset)

    def test_image_queryset_no_model_nor_device_firmware(self):
        (
            device,
            device_fw,
            img_org2,
            yuncore,
            mesh_image,
        ) = self._prepare_image_qs_test_env()
        # new DeviceFirmware, no model
        # returns all devices of the org
        device.model = ''
        device.save()
        form = DeviceFirmwareForm(device=device)
        self.assertEqual(form.fields['image'].queryset.count(), 3)
        self.assertIn(device_fw.image, form.fields['image'].queryset)
        self.assertIn(mesh_image, form.fields['image'].queryset)
        self.assertIn(yuncore, form.fields['image'].queryset)
        self.assertNotIn(img_org2, form.fields['image'].queryset)

    def test_admin_menu_groups(self):
        # Test menu group (openwisp-utils menu group) for Build, Category,
        # BatchUpgradeOperation models
        self._login()
        models = ['build', 'category', 'batchupgradeoperation']
        response = self.client.get(reverse('admin:index'))
        for model in models:
            with self.subTest(f'test menu group link {model} model'):
                url = reverse(f'admin:{self.app_label}_{model}_changelist')
                self.assertContains(response, f'class="mg-link" href="{url}"')
        with self.subTest('test firmware group is registered'):
            self.assertContains(
                response,
                '<div class="mg-dropdown-label">Firmware </div>',
                html=True,
            )

    def test_save_device_with_deleted_devicefirmware(self):
        self._login()
        device_fw = self._create_device_firmware()
        device = device_fw.device
        device_conn = device.deviceconnection_set.first()
        device_params = self._device_params.copy()
        device_params.update(
            {
                'devicefirmware-0-image': str(device_fw.image_id),
                'devicefirmware-0-id': str(device_fw.id),
                'organization': str(device.organization.id),
                'config-0-id': str(device.config.pk),
                'config-0-device': str(device.id),
                'deviceconnection_set-0-credentials': str(device_conn.credentials_id),
                'deviceconnection_set-0-id': str(device_conn.id),
                'devicefirmware-TOTAL_FORMS': 1,
                'devicefirmware-INITIAL_FORMS': 1,
            }
        )
        FirmwareImage.objects.all().delete()
        response = self.client.post(
            reverse('admin:config_device_change', args=[device.id]),
            data=device_params,
            follow=True,
        )
        self.assertEqual(response.status_code, 200)

    @capture_stderr()
    @mock.patch(
        'openwisp_firmware_upgrader.utils.get_upgrader_class_from_device_connection'
    )
    def test_device_firmware_upgrade_without_device_connection(
        self, captured_stderr, mocked_func, *args
    ):
        self._login()
        device_fw = self._create_device_firmware()
        device = device_fw.device
        device.deviceconnection_set.all().delete()
        response = self.client.get(
            reverse('admin:config_device_change', args=[device.id])
        )
        self.assertNotIn(
            '\'NoneType\' object has no attribute \'update_strategy\'',
            captured_stderr.getvalue(),
        )
        mocked_func.assert_not_called()
        self.assertEqual(response.status_code, 200)

    def test_deactivated_firmware_image_inline(self):
        self._login()
        device = self._create_config(organization=self._get_org()).device
        device.deactivate()
        response = self.client.get(
            reverse('admin:config_device_change', args=[device.id])
        )
        # Check that it is not possible to add a DeviceFirmwareImage to a
        # deactivated device in the admin interface.
        self.assertContains(
            response,
            '<input type="hidden" name="devicefirmware-MAX_NUM_FORMS"'
            ' value="0" id="id_devicefirmware-MAX_NUM_FORMS">',
        )
        self._create_device_firmware(device=device)
        response = self.client.get(
            reverse('admin:config_device_change', args=[device.id])
        )
        # Ensure that a deactivated device's existing DeviceFirmwareImage
        # is displayed as readonly in the admin interface.
        self.assertContains(
            response,
            '<div class="readonly">Test Category v0.1:'
            ' TP-Link WDR4300 v1 (OpenWrt 19.07 and later)</div>',
        )
        self.assertNotContains(
            response,
            '<select name="devicefirmware-0-image" id="id_devicefirmware-0-image">',
        )


_mock_updrade = 'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt.upgrade'
_mock_connect = 'openwisp_controller.connection.models.DeviceConnection.connect'


@mock.patch(_mock_updrade, return_value=True)
@mock.patch(_mock_connect, return_value=True)
class TestAdminTransaction(
    BaseTestAdmin, AdminActionPermTestMixin, TransactionTestCase
):
    def test_upgrade_selected_action_perms(self, *args):
        env = self._create_upgrade_env()
        org = env['d1'].organization
        self._create_firmwareless_device(organization=org)
        user = self._create_user(is_staff=True)
        self._create_org_user(user=user, organization=org, is_admin=True)
        # The user is redirected to the BatchUpgradeOperation page after success operation.
        # Thus, we need to add the permission to the user.
        user.user_permissions.add(
            Permission.objects.get(
                codename=f'change_{BatchUpgradeOperation._meta.model_name}'
            )
        )
        self._test_action_permission(
            path=self.build_list_url,
            action='upgrade_selected',
            user=user,
            obj=env['build1'],
            message=(
                'You can track the progress of this mass upgrade operation '
                'in this page. Refresh the page from time to time to check '
                'its progress.'
            ),
            required_perms=['change'],
            extra_payload={
                'upgrade_all': 'upgrade_all',
                'upgrade_options': '{"c": true}',
            },
        )

    def test_upgrade_related(self, *args):
        self._login()
        env = self._create_upgrade_env()
        self._create_firmwareless_device(organization=env['d1'].organization)
        # check state is good before proceeding
        fw = DeviceFirmware.objects.filter(
            image__build_id=env['build2'].pk
        ).select_related('image')
        self.assertEqual(Device.objects.count(), 3)
        self.assertEqual(UpgradeOperation.objects.count(), 0)
        self.assertEqual(fw.count(), 0)

        with self.subTest('Invalid upgrade_options'):
            response = self.client.post(
                self.build_list_url,
                {
                    'action': 'upgrade_selected',
                    'upgrade_related': 'upgrade_related',
                    'upgrade_options': 'invalid',
                    ACTION_CHECKBOX_NAME: (env['build2'].pk,),
                },
                follow=True,
            )
            self.assertContains(
                response, '<ul class="errorlist"><li>Enter a valid JSON.</li></ul>'
            )

        with self.subTest('Test with valid upgrade_options'):
            r = self.client.post(
                self.build_list_url,
                {
                    'action': 'upgrade_selected',
                    'upgrade_related': 'upgrade_related',
                    'upgrade_options': '{"c": true}',
                    ACTION_CHECKBOX_NAME: (env['build2'].pk,),
                },
                follow=True,
            )
            self.assertContains(r, '<li class="success">')
            self.assertContains(r, 'track the progress')
            self.assertEqual(
                UpgradeOperation.objects.filter(upgrade_options={'c': True}).count(), 2
            )
            self.assertEqual(fw.count(), 2)

    def test_upgrade_all(self, *args):
        self._login()
        env = self._create_upgrade_env()
        self._create_firmwareless_device(organization=env['d1'].organization)
        # check state is good before proceeding
        fw = DeviceFirmware.objects.filter(
            image__build_id=env['build2'].pk
        ).select_related('image')
        self.assertEqual(Device.objects.count(), 3)
        self.assertEqual(UpgradeOperation.objects.count(), 0)
        self.assertEqual(fw.count(), 0)

        with self.subTest('Invalid upgrade_options'):
            response = self.client.post(
                self.build_list_url,
                {
                    'action': 'upgrade_selected',
                    'upgrade_all': 'upgrade_all',
                    'upgrade_options': 'invalid',
                    ACTION_CHECKBOX_NAME: (env['build2'].pk,),
                },
                follow=True,
            )
            self.assertEqual(response.status_code, 200)
            self.assertContains(
                response, '<ul class="errorlist"><li>Enter a valid JSON.</li></ul>'
            )

        with self.subTest('Test with valid upgrade_options'):
            response = self.client.post(
                self.build_list_url,
                {
                    'action': 'upgrade_selected',
                    'upgrade_all': 'upgrade_all',
                    'upgrade_options': '{"c": true}',
                    ACTION_CHECKBOX_NAME: (env['build2'].pk,),
                },
                follow=True,
            )
            self.assertContains(response, '<li class="success">')
            self.assertContains(response, 'track the progress')
            self.assertEqual(
                UpgradeOperation.objects.filter(upgrade_options={'c': True}).count(), 3
            )
            self.assertEqual(fw.count(), 3)
            self.assertContains(
                response,
                (
                    '<div class="readonly"><ul class="readonly-upgrade-options">'
                    '<li><img src="/static/admin/img/icon-yes.svg" alt="yes">'
                    'Attempt to preserve all changed files in /etc/ (-c)</li>'
                    '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                    'Attempt to preserve all changed files in /, except those from '
                    'packages but including changed confs. (-o)</li>'
                    '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                    'Do not save configuration over reflash (-n)</li>'
                    '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                    'Skip from backup files that are equal to those in /rom (-u)</li>'
                    '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                    'Do not attempt to restore the partition table after flash. (-p)</li>'
                    '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                    'Include in backup a list of current installed packages at '
                    '/etc/backup/installed_packages.txt (-k)</li>'
                    '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                    'Flash image even if image checks fail, this is dangerous! (-F)</li></ul></div>'
                ),
                html=True,
            )

    def test_massive_upgrade_operation_page(self, *args):
        self.test_upgrade_all()
        uo = UpgradeOperation.objects.first()
        url = reverse(
            f'admin:{self.app_label}_batchupgradeoperation_change', args=[uo.batch.pk]
        )
        response = self.client.get(url)
        self.assertContains(response, 'Success rate')
        self.assertContains(response, 'Failure rate')
        self.assertContains(response, 'Abortion rate')

    def test_recent_upgrades(self, *args):
        self._login()
        env = self._create_upgrade_env()
        url = reverse('admin:config_device_change', args=[env['d2'].pk])
        r = self.client.get(url)
        self.assertNotContains(r, 'Recent Firmware Upgrades')
        env['build2'].batch_upgrade(firmwareless=True)
        r = self.client.get(url)
        self.assertContains(r, 'Recent Firmware Upgrades')

    def test_upgrade_operation_inline(self, *args):
        device_fw = self._create_device_firmware()
        device_fw.save(upgrade=True)
        device = device_fw.device
        request = self.make_device_admin_request(device.pk)
        request.user = User.objects.first()
        deviceadmin = DeviceAdmin(model=Device, admin_site=admin.site)
        self.assertNotIn(
            DeviceUpgradeOperationInline, deviceadmin.get_inlines(request, obj=None)
        )
        self.assertIn(
            DeviceUpgradeOperationInline, deviceadmin.get_inlines(request, obj=device)
        )

    def test_upgrade_operation_inline_queryset(self, *args):
        device_fw = self._create_device_firmware()
        device_fw.save(upgrade=True)
        # expect only 1
        uo = device_fw.device.upgradeoperation_set.get()
        device = device_fw.device
        request = self.make_device_admin_request(device.pk)
        request.user = User.objects.first()
        inline = DeviceUpgradeOperationInline(Device, admin.site)
        qs = inline.get_queryset(request)
        self.assertEqual(qs.count(), 1)
        self.assertIn(uo, qs)
        uo.created = localtime() - timedelta(days=30)
        uo.modified = uo.created
        uo.save()
        qs = inline.get_queryset(request)
        self.assertEqual(qs.count(), 0)

    def test_device_firmware_upgrade_options(self, *args):
        self._login()
        device_fw = self._create_device_firmware()
        device = device_fw.device
        device_conn = device.deviceconnection_set.first()
        device_params = self._device_params.copy()
        build = self._create_build(version='0.2')
        image = self._create_firmware_image(build=build)
        upgrade_options = {
            'c': True,
            'o': False,
            'u': False,
            'n': False,
            'p': False,
            'k': False,
            'F': True,
        }
        device_params.update(
            {
                'model': device.model,
                'devicefirmware-0-image': str(image.id),
                'devicefirmware-0-id': str(device_fw.id),
                'devicefirmware-0-upgrade_options': json.dumps(upgrade_options),
                'organization': str(device.organization.id),
                'config-0-id': str(device.config.pk),
                'config-0-device': str(device.id),
                'deviceconnection_set-0-credentials': str(device_conn.credentials_id),
                'deviceconnection_set-0-id': str(device_conn.id),
                'deviceconnection_set-0-update_strategy': conn_settings.DEFAULT_UPDATE_STRATEGIES[
                    0
                ][
                    0
                ],
                'deviceconnection_set-0-enabled': True,
                'devicefirmware-TOTAL_FORMS': 1,
                'devicefirmware-INITIAL_FORMS': 1,
                'upgradeoperation_set-TOTAL_FORMS': 0,
                'upgradeoperation_set-INITIAL_FORMS': 0,
                'upgradeoperation_set-MIN_NUM_FORMS': 0,
                'upgradeoperation_set-MAX_NUM_FORMS': 0,
                '_continue': True,
            }
        )
        response = self.client.post(
            reverse('admin:config_device_change', args=[device.id]),
            data=device_params,
            follow=True,
        )
        self.assertEqual(response.status_code, 200)
        self.assertEqual(device.upgradeoperation_set.count(), 1)
        upgrade_operation = device.upgradeoperation_set.first()
        self.assertEqual(upgrade_operation.upgrade_options, upgrade_options)
        self.assertContains(
            response,
            (
                '<div class="readonly"><ul class="readonly-upgrade-options"><li>'
                '<img src="/static/admin/img/icon-yes.svg" alt="yes">'
                'Attempt to preserve all changed files in /etc/ (-c)</li>'
                '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                'Attempt to preserve all changed files in /, except those from packages '
                'but including changed confs. (-o)</li>'
                '<li><img src="/static/admin/img/icon-no.svg" '
                'alt="no">Do not save configuration over reflash (-n)</li>'
                '<li><img src="/static/admin/img/icon-no.svg" alt="no">Skip from backup files '
                'that are equal to those in /rom (-u)</li>'
                '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                'Do not attempt to restore the partition table after flash. (-p)</li>'
                '<li><img src="/static/admin/img/icon-no.svg" alt="no">'
                'Include in backup a list of current installed packages at '
                '/etc/backup/installed_packages.txt (-k)</li>'
                '<li><img src="/static/admin/img/icon-yes.svg" alt="yes">'
                'Flash image even if image checks fail, this is dangerous! (-F)</li></ul></div>'
            ),
            html=True,
        )

    @mock.patch.object(OpenWisp1, 'SCHEMA', None)
    def test_using_upgrade_options_with_unsupported_upgrader(self, *args):
        self._login()
        device_fw = self._create_device_firmware()
        device = device_fw.device
        device.config.backend = 'netjsonconfig.OpenWisp'
        device.config.save()
        device_conn = device.deviceconnection_set.first()
        device_conn.update_strategy = conn_settings.DEFAULT_UPDATE_STRATEGIES[1][0]
        device_conn.save()
        device_params = self._device_params.copy()
        build = self._create_build(version='0.2')
        image = self._create_firmware_image(build=build)
        upgrade_options = {
            'c': True,
            'o': False,
            'u': False,
            'n': False,
            'p': False,
            'k': False,
            'F': True,
        }
        device_params.update(
            {
                'model': device.model,
                'devicefirmware-0-image': str(image.id),
                'devicefirmware-0-id': str(device_fw.id),
                'devicefirmware-0-upgrade_options': json.dumps(upgrade_options),
                'organization': str(device.organization.id),
                'config-0-id': str(device.config.pk),
                'config-0-device': str(device.id),
                'deviceconnection_set-0-credentials': str(device_conn.credentials_id),
                'deviceconnection_set-0-id': str(device_conn.id),
                'deviceconnection_set-0-update_strategy': (
                    conn_settings.DEFAULT_UPDATE_STRATEGIES[1][0]
                ),
                'deviceconnection_set-0-enabled': True,
                'devicefirmware-TOTAL_FORMS': 1,
                'devicefirmware-INITIAL_FORMS': 1,
                'upgradeoperation_set-TOTAL_FORMS': 0,
                'upgradeoperation_set-INITIAL_FORMS': 0,
                'upgradeoperation_set-MIN_NUM_FORMS': 0,
                'upgradeoperation_set-MAX_NUM_FORMS': 0,
                '_continue': True,
            }
        )

        with self.subTest('Test DeviceFirmwareInline does not have schema defined'):
            response = self.client.get(
                reverse('admin:config_device_change', args=[device.id])
            )
            self.assertContains(
                response, '<script>\nvar firmwareUpgraderSchema = null\n</script>'
            )

        with self.subTest('Test using upgrade options with unsupported upgrader'):
            response = self.client.post(
                reverse('admin:config_device_change', args=[device.id]),
                data=device_params,
                follow=True,
            )
            self.assertContains(
                response,
                (
                    '<ul class="errorlist nonfield"><li>Using upgrade '
                    'options is not allowed with this upgrader.</li></ul>'
                ),
            )

        with self.subTest('Test upgrading without upgrade options'):
            del device_params['devicefirmware-0-upgrade_options']
            response = self.client.post(
                reverse('admin:config_device_change', args=[device.id]),
                data=device_params,
                follow=True,
            )
            self.assertContains(
                response,
                (
                    '<div class="readonly">Upgrade options are '
                    'not supported for this upgrader.</div>'
                ),
            )


del TestConfigAdmin



import logging

from django.utils.module_loading import import_string

from . import settings as app_settings

logger = logging.getLogger(__name__)


def get_upgrader_schema_for_device(device):
    upgrader_class = get_upgrader_class_for_device(device)
    return getattr(upgrader_class, 'SCHEMA', None)


def get_upgrader_class_for_device(device):
    """
    Returns firmware upgrader class for a device depending
    on update_strategy of device's DeviceConnection.

    It only takes the first DeviceConnection object into consideration.
    This function makes the following assumptions:
        - a device cannot have DeviceConnection objects of
          two different update_strategy.
        - an upgrade cannot be performed on a device without a
          device connection
    """
    device_conn = device.deviceconnection_set.filter(
        update_strategy__icontains='ssh',
        enabled=True,
    ).first()
    if not device_conn:
        raise device.deviceconnection_set.model.DoesNotExist
    return get_upgrader_class_from_device_connection(device_conn)


def get_upgrader_class_from_device_connection(device_conn):
    try:
        upgrader_class = app_settings.UPGRADERS_MAP[device_conn.update_strategy]
        upgrader_class = import_string(upgrader_class)
    except (AttributeError, ImportError, KeyError) as e:
        logger.exception(e)
        return
    return upgrader_class



from unittest import mock

from celery.exceptions import SoftTimeLimitExceeded
from django.test import TransactionTestCase

from openwisp_utils.tests import capture_any_output

from .. import tasks
from ..swapper import load_model
from .base import TestUpgraderMixin

BatchUpgradeOperation = load_model('BatchUpgradeOperation')
UpgradeOperation = load_model('UpgradeOperation')


class TestTasks(TestUpgraderMixin, TransactionTestCase):
    _mock_upgrade = 'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt.upgrade'
    _mock_connect = 'openwisp_controller.connection.models.DeviceConnection.connect'

    @mock.patch(_mock_upgrade, side_effect=SoftTimeLimitExceeded())
    @mock.patch(_mock_connect, return_value=True)
    @mock.patch(
        'openwisp_firmware_upgrader.base.models.AbstractUpgradeOperation.upgrade',
        side_effect=SoftTimeLimitExceeded(),
    )
    @capture_any_output()
    def test_upgrade_firmware_timeout(self, *args):
        device_fw = self._create_device_firmware(upgrade=True)
        self.assertEqual(UpgradeOperation.objects.count(), 1)
        uo = device_fw.image.upgradeoperation_set.first()
        self.assertEqual(uo.status, 'failed')
        self.assertIn('Operation timed out.', uo.log)

    @mock.patch(_mock_upgrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    @mock.patch(
        'openwisp_firmware_upgrader.base.models.AbstractDeviceFirmware.create_upgrade_operation',
        side_effect=SoftTimeLimitExceeded(),
    )
    @capture_any_output()
    def test_batch_upgrade_timeout(self, *args):
        env = self._create_upgrade_env()
        batch = BatchUpgradeOperation.objects.create(build=env['build2'])
        # will be executed synchronously due to CELERY_IS_EAGER = True
        tasks.batch_upgrade_operation.delay(batch_id=batch.pk, firmwareless=False)
        self.assertEqual(BatchUpgradeOperation.objects.count(), 1)
        batch = BatchUpgradeOperation.objects.first()
        self.assertEqual(batch.status, 'failed')

    @mock.patch(_mock_upgrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    @mock.patch('logging.Logger.warning')
    def test_upgrade_firmware_resilience(self, mocked_logger, *args):
        upgrade_op_id = UpgradeOperation().id
        tasks.upgrade_firmware.run(upgrade_op_id)
        mocked_logger.assert_called_with(
            f'The UpgradeOperation object with id {upgrade_op_id} has been deleted'
        )

    @mock.patch(_mock_upgrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    @mock.patch('logging.Logger.warning')
    def test_batch_upgrade_operation_resilience(self, mocked_logger, *args):
        batch_id = BatchUpgradeOperation().id
        tasks.batch_upgrade_operation.run(batch_id=batch_id, firmwareless=False)
        mocked_logger.assert_called_with(
            f'The BatchUpgradeOperation object with id {batch_id} has been deleted'
        )



"""
Mapping between hardware models and firwmare image
This if focused on OpenWrt only for now, but it should
be possible to add support for different embedded
systems in the future.
"""
from collections import OrderedDict

from . import settings as app_settings

if app_settings.CUSTOM_OPENWRT_IMAGES:
    OPENWRT_FIRMWARE_IMAGE_MAP = OrderedDict(app_settings.CUSTOM_OPENWRT_IMAGES)
else:  # pragma: no cover
    OPENWRT_FIRMWARE_IMAGE_MAP = OrderedDict()

OPENWRT_FIRMWARE_IMAGE_MAP.update(
    OrderedDict(
        (
            (
                'ar71xx-generic-cf-e320n-v2-squashfs-sysupgrade.bin',
                {
                    'label': 'COMFAST CF-E320N v2 (OpenWrt 19.07 and earlier)',
                    'boards': ('COMFAST CF-E320N v2',),
                },
            ),
            (
                'ath79-generic-comfast_cf-e375ac-squashfs-sysupgrade.bin',
                {
                    'label': 'COMFAST CF-E375AC',
                    'boards': ('COMFAST CF-E375AC',),
                },
            ),
            (
                'dongwon_dw02-412h-128m-squashfs-sysupgrade.bin',
                {
                    'label': 'Dongwon T&I DW02-412H (128M) / KT GiGA WiFi home (128M)',
                    'boards': ('DW02-412H-128M-NAND',),
                },
            ),
            (
                'ipq807x-generic-edgecore_eap102-squashfs-sysupgrade.bin',
                {
                    'label': 'Edgecore EAP102',
                    'boards': ('Edgecore EAP102',),
                },
            ),
            (
                'ipq40xx-generic-engenius_eap1300-squashfs-sysupgrade.bin',
                {
                    'label': 'EnGenius EAP1300',
                    'boards': ('EnGenius EAP1300',),
                },
            ),
            (
                'ath79-nand-glinet_gl-ar300m-nand-squashfs-sysupgrade.bin',
                {
                    'label': 'GL.iNet GL-AR300M (NAND)',
                    'boards': ('GL.iNet GL-AR300M (NAND)',),
                },
            ),
            (
                'ramips-mt76x8-gl-mt300n-v2-squashfs-sysupgrade.bin',
                {'label': 'GL.iNet GL-MT300N-V2', 'boards': ('GL-MT300N-V2',)},
            ),
            (
                'mvebu-cortexa9-linksys_wrt1900acs-squashfs-sysupgrade.img',
                {'label': 'Linksys WRT1900ACS', 'boards': ('Linksys WRT1900ACS',)},
            ),
            (
                'mvebu-cortexa9-linksys_wrt3200acm-squashfs-sysupgrade.img',
                {'label': 'Linksys WRT3200ACM', 'boards': ('Linksys WRT3200ACM',)},
            ),
            (
                'ipq40xx-mikrotik-mikrotik_wap-ac-squashfs-sysupgrade.bin',
                {'label': 'MikroTik wAP ac', 'boards': ('MikroTik wAP ac',)},
            ),
            (
                'brcm2708-bcm2709-rpi-2-ext4-sysupgrade.img.gz',
                {
                    'label': 'Raspberry Pi 2 Model B',
                    'boards': (
                        'Raspberry Pi 2 Model B Rev 1.0',
                        'Raspberry Pi 2 Model B Rev 1.1',
                        'Raspberry Pi 2 Model B Rev 1.2',
                    ),
                },
            ),
            (
                'brcm2708-bcm2710-rpi-3-ext4-sysupgrade.img.gz',
                {
                    'label': 'Raspberry Pi 3 Model B',
                    'boards': ('Raspberry Pi 3 Model B Rev 1.2',),
                },
            ),
            (
                'ar71xx-generic-archer-c7-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v1 (OpenWrt 19.07 and earlier)',
                    'boards': ('tplink,archer-c7-v1',),
                },
            ),
            (
                'ath79-generic-tplink_archer-c7-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v1 (OpenWrt 19.07 and later)',
                    'boards': ('tplink,archer-c7-v1',),
                },
            ),
            (
                'ar71xx-generic-archer-c7-v2-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v2 (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-Link Archer C7 v2', 'TP-Link Archer C7 v3'),
                },
            ),
            (
                'ath79-generic-tplink_archer-c7-v2-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v2 (OpenWrt 19.07 and later)',
                    'boards': ('TP-Link Archer C7 v2', 'TP-Link Archer C7 v3'),
                },
            ),
            (
                'ar71xx-generic-archer-c7-v4-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v4 (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-Link Archer C7 v4',),
                },
            ),
            (
                'ath79-generic-tplink_archer-c7-v4-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v4 (OpenWrt 19.07 and later)',
                    'boards': ('TP-Link Archer C7 v4',),
                },
            ),
            (
                'ar71xx-generic-archer-c7-v5-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v5 (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-Link Archer C7 v5',),
                },
            ),
            (
                'ath79-generic-tplink_archer-c7-v5-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C7 v5 (OpenWrt 19.07 and later)',
                    'boards': ('TP-Link Archer C7 v5',),
                },
            ),
            (
                'ramips-mt76x8-tplink_archer-c50-v4-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link Archer C50 v4',
                    'boards': ('TP-Link Archer C50 v4',),
                },
            ),
            (
                'ar71xx-generic-cpe210-220-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-LINK CPE210 v3 (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-LINK CPE210 v1', 'TP-LINK CPE220 v1'),
                },
            ),
            (
                'ath79-generic-tplink_cpe210-v2-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-LINK CPE210 v2 (OpenWrt 19.07 and later)',
                    'boards': ('TP-LINK CPE210 v2',),
                },
            ),
            (
                'ath79-generic-tplink_cpe210-v3-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-LINK CPE210 v3 (OpenWrt 19.07 and later)',
                    'boards': ('TP-LINK CPE210 v3',),
                },
            ),
            (
                'ath79-generic-tplink_cpe510-v3-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-LINK CPE510 v3 (OpenWrt 19.07 and later)',
                    'boards': ('TP-LINK CPE510 v3',),
                },
            ),
            (
                'ath79-generic-tplink_eap225-outdoor-v3-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link EAP225-Outdoor v3',
                    'boards': ('TP-Link EAP225-Outdoor v3',),
                },
            ),
            (
                'ar71xx-generic-tl-wdr3600-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WDR3600 v1 (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-Link TL-WDR3600 v1',),
                },
            ),
            (
                'ath79-generic-tplink_tl-wdr3600-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WDR3600 v1 (OpenWrt 19.07 and later)',
                    'boards': ('TP-Link TL-WDR3600 v1',),
                },
            ),
            (
                'ar71xx-generic-tl-wdr4300-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WDR4300 v1 (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-Link TL-WDR4300 v1',),
                },
            ),
            (
                'ath79-generic-tplink_tl-wdr4300-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WDR4300 v1 (OpenWrt 19.07 and later)',
                    'boards': ('TP-Link TL-WDR4300 v1',),
                },
            ),
            (
                'ar71xx-generic-tl-wdr4300-v1-il-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WDR4300 v1 Israel Version (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-LINK TL-WDR4300 v1 (IL)',),
                },
            ),
            (
                'ath79-generic-tplink_tl-wdr4300-v1-il-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WDR4300 v1 Israel Version (OpenWrt 19.07 and later)',
                    'boards': ('TP-LINK TL-WDR4300 v1 (IL)',),
                },
            ),
            (
                'ar71xx-generic-tl-wr2543-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WR2543N/ND (OpenWrt 19.07 and earlier)',
                    'boards': ('TP-Link TL-WR2543N/ND',),
                },
            ),
            (
                'ath79-generic-tplink_tl-wr2543-v1-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link WR2543N/ND (OpenWrt 19.07 and later)',
                    'boards': ('TP-Link TL-WR2543N/ND',),
                },
            ),
            (
                'ramips-mt76x8-tplink_tl-wr902ac-v3-squashfs-sysupgrade.bin',
                {
                    'label': 'TP-Link TL-WR902AC v3',
                    'boards': ('TP-Link TL-WR902AC v3',),
                },
            ),
            (
                'ar71xx-generic-ubnt-airrouter-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti AirRouter (OpenWrt 19.07 and earlier)',
                    'boards': ('Ubiquiti AirRouter',),
                },
            ),
            (
                'ath79-generic-ubnt_airrouter-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti AirRouter (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti AirRouter',),
                },
            ),
            (
                'octeon-erlite-squashfs-sysupgrade.tar',
                {
                    'label': 'Ubiquiti EdgeRouter Lite',
                    'boards': ('Ubiquiti EdgeRouter Lite',),
                },
            ),
            # Nanostation Loco M XW AR71XX
            (
                'ar71xx-generic-ubnt-loco-m-xw-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation Loco M2 - XW (OpenWrt 19.07 and earlier)',
                    'boards': ('Ubiquiti Loco XW',),
                },
            ),
            # Nanostation Loco M XM ATH79
            (
                'ath79-generic-ubnt_nanostation-loco-m-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation Loco M (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti Nanostation Loco M',),
                },
            ),
            # Nanostation Loco M XW ATH79
            (
                'ath79-generic-ubnt_nanostation-loco-m-xw-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation Loco M - XW (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti Nanostation Loco M (XW)',),
                },
            ),
            # Nanostation M XW AR71XX
            (
                'ar71xx-generic-ubnt-nano-m-xw-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation M - XW (OpenWrt 19.07 and earlier)',
                    'boards': ('Ubiquiti Nano M XW',),
                },
            ),
            # Nanostation M XM AR71XX
            (
                'ar71xx-generic-ubnt-nano-m-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation M (OpenWrt 19.07 and earlier)',
                    'boards': (
                        'Ubiquiti Nano-M',
                        'Ubiquiti NanoStation M2',
                        'Ubiquiti NanoStation M5',
                        'Ubiquiti NanoStation loco M2',
                        'Ubiquiti NanoStation loco M5',
                    ),
                },
            ),
            # Nanostation M XW ATH79
            (
                'ath79-generic-ubnt_nanostation-m-xw-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation M - XW (OpenWrt 19.07 and later)',
                    'boards': (
                        'Ubiquiti Nanostation M (XW)',
                        'Ubiquiti Nanostation M XW',
                    ),
                },
            ),
            # Nanostation M XM ATH79
            (
                'ath79-generic-ubnt_nanostation-m-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Nanostation M (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti Nanostation M',),
                },
            ),
            # Bullet XW AR71XX
            (
                'ar71xx-generic-ubnt-bullet-m-xw-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Picostation Bullet XW (OpenWrt 19.07 and earlier)',
                    'boards': ('Ubiquiti Bullet-M XW',),
                },
            ),
            # Picostation M2HP & Bullet AR71XX
            (
                'ar71xx-generic-ubnt-bullet-m-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Picostation M2HP & Bullet (OpenWrt 19.07 and earlier)',
                    'boards': (
                        'Ubiquiti Bullet-M',
                        'Ubiquiti PicoStation M2',
                        'Ubiquiti PicoStation M2HP',
                    ),
                },
            ),
            # Picostation M ATH79
            (
                'ath79-generic-ubnt_picostation-m-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Picostation M (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti Picostation M',),
                },
            ),
            # Ubiquiti Unifi
            (
                'ath79-generic-ubnt_unifi-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti UniFi (OpenWRT 19.07 and later)',
                    'boards': ('Ubiquiti UniFi', 'Ubiquiti UniFi AP'),
                },
            ),
            # Ubiquiti UniFi AC LR ATH79
            (
                'ath79-generic-ubnt_unifiac-lr-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti UniFi AC LR (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti UniFi AC LR',),
                },
            ),
            # Unifi AC Mesh AR71XX
            (
                'ar71xx-generic-ubnt-unifiac-mesh-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Unifi AC Mesh (OpenWrt 19.07 and earlier)',
                    'boards': (
                        'Ubiquiti UniFi AC-Mesh',
                        'Ubiquiti UniFi-AC-MESH',
                        'Ubiquiti UniFi-AC-LITE/MESH',
                    ),
                },
            ),
            # Unifi AC Mesh ATH79
            (
                'ath79-generic-ubnt_unifiac-mesh-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Unifi AC Mesh (OpenWrt 19.07 and later)',
                    'boards': (
                        'Ubiquiti UniFi AC Mesh',
                        'Ubiquiti UniFi AC-Mesh',
                        'Ubiquiti UniFi-AC-MESH',
                        'Ubiquiti UniFi-AC-LITE/MESH',
                    ),
                },
            ),
            # Unifi AC Mesh Pro AR71XX
            (
                'ar71xx-generic-ubnt-unifiac-mesh-pro-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Unifi AC Mesh-Pro (OpenWrt 19.07 and earlier)',
                    'boards': ('Ubiquiti UniFi AC-Mesh-Pro',),
                },
            ),
            # Unifi AC Mesh Pro ATH79
            (
                'ath79-generic-ubnt_unifiac-mesh-pro-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti Unifi AC Mesh-Pro (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti UniFi AC-Mesh Pro',),
                },
            ),
            # Unifi AC Pro ATH79
            (
                'ath79-generic-ubnt_unifiac-pro-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti UniFi AC Pro (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti UniFi AC Pro', 'Ubiquiti UniFi-AC-PRO'),
                },
            ),
            # Unifi AP Pro ATH79
            (
                'ath79-generic-ubnt_unifi-ap-pro-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti UniFi AP Pro (OpenWrt 19.07 and later)',
                    'boards': ('Ubiquiti UniFi AP Pro',),
                },
            ),
            # Unifi AP Pro AR71XX
            (
                'ar71xx-generic-ubnt-uap-pro-squashfs-sysupgrade.bin',
                {
                    'label': 'Ubiquiti UniFi AP Pro (OpenWrt 19.07 and earlier)',
                    'boards': ('Ubiquiti UAP Pro',),
                },
            ),
            (
                'ramips-mt7621-yuncore_ax820-squashfs-sysupgrade.bin',
                {
                    'label': 'YunCore AX820',
                    'boards': ('YunCore AX820',),
                },
            ),
            (
                'ramips-mt76x8-yuncore_cpe200-squashfs-sysupgrade.bin',
                {
                    'label': 'YunCore CPE200',
                    'boards': ('Yuncore CPE200',),
                },
            ),
            (
                'ramips-mt7621-yuncore_g720-squashfs-sysupgrade.bin',
                {
                    'label': 'YunCore G720',
                    'boards': ('YunCore G720',),
                },
            ),
            (
                'ramips-mt76x8-yuncore_m300-squashfs-sysupgrade.bin',
                {
                    'label': 'YunCore M300',
                    'boards': ('Yuncore M300',),
                },
            ),
            (
                'ar71xx-generic-xd3200-squashfs-sysupgrade.bin',
                {
                    'label': 'YunCore XD3200 (OpenWrt 19.07 and earlier)',
                    'boards': ('YunCore XD3200',),
                },
            ),
            (
                'ramips-mt7620-zbtlink_zbt-we1026-5g-16m-squashfs-sysupgrade.bin',
                {
                    'label': 'Zbtlink ZBT-WE1026-5G (16M)',
                    'boards': ('Zbtlink ZBT-WE1026-5G (16M)',),
                },
            ),
            (
                'ramips-mt7620-zbtlink_zbt-we826-16m-squashfs-sysupgrade.bin',
                {
                    'label': 'Zbtlink ZBT-WE826 (16M)',
                    'boards': ('Zbtlink ZBT-WE826 (16M)',),
                },
            ),
            (
                'ramips-mt7620-zbtlink_zbt-we826-32m-squashfs-sysupgrade.bin',
                {
                    'label': 'Zbtlink ZBT-WE826 (32M)',
                    'boards': ('Zbtlink ZBT-WE826 (32M)',),
                },
            ),
            (
                'ramips-mt7621-zbt-wg3526-16M-squashfs-sysupgrade.bin',
                {
                    'label': 'Zbtlink ZBT-WG3526 (16M)',
                    'boards': ('ZBT-WG3526 (16M)', 'Zbtlink ZBT-WG3526 (16M)'),
                },
            ),
            (
                'ramips-mt7621-zbt-wg3526-32M-squashfs-sysupgrade.bin',
                {
                    'label': 'Zbtlink ZBT-WG3526 (32M)',
                    'boards': ('ZBT-WG3526 (32M)', 'Zbtlink ZBT-WG3526 (32M)'),
                },
            ),
            (
                'x86-64-generic-squashfs-combined.img.gz',
                {
                    'label': 'Generic x86/64 (QEMU/KVM)',
                    'boards': (
                        'x86_64',
                        'PC Engines APU2',
                        'PC Engines apu2',
                        'PC Engines apu6',
                    ),
                },
            ),
            (
                'x86-64-combined-squashfs.img.gz',
                {
                    'label': 'VMware, Inc. VMware Virtual Platform',
                    'boards': ('VMware, Inc. VMware Virtual Platform',),
                },
            ),
            (
                'x86-generic-combined-squashfs.img.gz',
                {
                    'label': 'Generic x86/32 bit',
                    'boards': ('x86',),
                },
            ),
            (
                'x86-geode-combined-squashfs.img.gz',
                {
                    'label': 'x86 Geode(TM) Integrated Processor by AMD',
                    'boards': ('Geode(TM) Integrated Processor by AMD PCS', 'Alix 2D2'),
                },
            ),
        )
    )
)

# OpenWrt only for now, in the future we'll merge
# different dictionaries representing different firmwares
# eg: AirOS, Raspbian
FIRMWARE_IMAGE_MAP = OPENWRT_FIRMWARE_IMAGE_MAP

# Allows getting type from image board
REVERSE_FIRMWARE_IMAGE_MAP = {}
# Choices used in model
FIRMWARE_IMAGE_TYPE_CHOICES = []

for key, info in FIRMWARE_IMAGE_MAP.items():
    FIRMWARE_IMAGE_TYPE_CHOICES.append((key, info['label']))
    for board in info['boards']:
        REVERSE_FIRMWARE_IMAGE_MAP[board] = key



import io
from contextlib import redirect_stdout
from unittest import mock

import swapper
from celery.exceptions import Retry
from django.core.exceptions import ValidationError
from django.test import TestCase, TransactionTestCase

from openwisp_utils.tests import capture_any_output

from .. import settings as app_settings
from ..hardware import FIRMWARE_IMAGE_MAP, REVERSE_FIRMWARE_IMAGE_MAP
from ..swapper import load_model
from ..tasks import upgrade_firmware
from .base import TestUpgraderMixin

Group = swapper.load_model('openwisp_users', 'Group')
BatchUpgradeOperation = load_model('BatchUpgradeOperation')
Build = load_model('Build')
Category = load_model('Category')
DeviceFirmware = load_model('DeviceFirmware')
FirmwareImage = load_model('FirmwareImage')
UpgradeOperation = load_model('UpgradeOperation')
DeviceConnection = swapper.load_model('connection', 'DeviceConnection')
Credentials = swapper.load_model('connection', 'Credentials')
Device = swapper.load_model('config', 'Device')


class TestModels(TestUpgraderMixin, TestCase):
    app_label = 'openwisp_firmware_upgrader'
    os = 'OpenWrt 19.07-SNAPSHOT r11061-6ffd4d8a4d'
    image_type = REVERSE_FIRMWARE_IMAGE_MAP['YunCore XD3200']

    def test_category_str(self):
        c = Category(name='WiFi Hotspot')
        self.assertEqual(str(c), c.name)

    def test_build_str(self):
        c = self._create_category()
        b = Build(category=c, version='0.1')
        self.assertIn(c.name, str(b))
        self.assertIn(b.version, str(b))

    def test_build_str_no_category(self):
        b = Build()
        self.assertIsNotNone(str(b))

    def test_build_clean(self):
        org = self._get_org()
        cat1 = self._get_category(organization=org)
        cat2 = self._create_category(name='New category', organization=org)
        b1 = self._create_build(organization=org, category=cat1, os=self.os)

        with self.subTest('validation error should be raised'):
            try:
                self._create_build(organization=org, category=cat2, os=self.os)
            except ValidationError as e:
                self.assertIn('os', e.message_dict)
            else:
                self.fail('ValidationError not raised')

        with self.subTest('1 build object expected'):
            self.assertEqual(Build.objects.count(), 1)

        with self.subTest('validating the same object again should work'):
            b1.full_clean()

        with self.subTest('validation error should be raised on empty category'):
            try:
                b2 = self._create_build(
                    os=self.os + '_2', version='0.2', organization=org
                )
                b2.category = None
                b2.full_clean()
            except ValidationError as e:
                self.assertIn('category', e.message_dict)
            else:
                self.fail('ValidationError not raised when build category is empty')

    def test_fw_str(self):
        fw = self._create_firmware_image()
        self.assertIn(str(fw.build), str(fw))
        self.assertIn(fw.build.category.name, str(fw))

    def test_fw_str_new(self):
        fw = FirmwareImage()
        self.assertIsNotNone(str(fw))

    def test_fw_auto_type(self):
        fw = self._create_firmware_image(type='')
        self.assertEqual(fw.type, self.TPLINK_4300_IMAGE)

    def test_device_firmware_image_invalid_org(self):
        device_fw = self._create_device_firmware()
        org2 = self._create_org(name='org2')
        build2 = self._create_build(organization=org2)
        img2 = self._create_firmware_image(build=build2)

        device_fw.image = img2
        try:
            device_fw.full_clean()
        except ValidationError as e:
            self.assertIn('image', e.message_dict)
        else:
            self.fail('ValidationError not raised')

    def test_device_fw_image_changed(self, *args):
        with mock.patch(
            f'{self.app_label}.models.UpgradeOperation.upgrade', return_value=None
        ):
            device_fw = DeviceFirmware()
            self.assertIsNone(device_fw._old_image)
            # save
            device_fw = self._create_device_firmware(upgrade=False)
            self.assertEqual(device_fw._old_image, device_fw.image)
            self.assertEqual(UpgradeOperation.objects.count(), 0)
            # init
            device_fw = DeviceFirmware.objects.first()
            self.assertEqual(device_fw._old_image, device_fw.image)
            # change
            build2 = self._create_build(
                category=device_fw.image.build.category, version='0.2'
            )
            fw2 = self._create_firmware_image(build=build2, type=device_fw.image.type)
            old_image = device_fw.image
            device_fw.image = fw2
            self.assertNotEqual(device_fw._old_image, device_fw.image)
            self.assertEqual(device_fw._old_image, old_image)
            device_fw.full_clean()
            device_fw.save()
            self.assertEqual(UpgradeOperation.objects.count(), 1)
            self.assertEqual(BatchUpgradeOperation.objects.count(), 0)

    def test_device_fw_created(self, *args):
        with mock.patch(
            f'{self.app_label}.models.UpgradeOperation.upgrade', return_value=None
        ):
            self._create_device_firmware(upgrade=True)
            self.assertEqual(UpgradeOperation.objects.count(), 1)
            self.assertEqual(BatchUpgradeOperation.objects.count(), 0)

    def test_device_fw_image_saved_not_installed(self, *args):
        with mock.patch(
            f'{self.app_label}.models.UpgradeOperation.upgrade', return_value=None
        ):
            device_fw = DeviceFirmware()
            self.assertIsNone(device_fw._old_image)
            # save
            device_fw = self._create_device_firmware(upgrade=False, installed=False)
            self.assertEqual(device_fw._old_image, device_fw.image)
            self.assertEqual(UpgradeOperation.objects.count(), 0)
            device_fw.full_clean()
            device_fw.save()
            self.assertEqual(UpgradeOperation.objects.count(), 1)
            self.assertEqual(BatchUpgradeOperation.objects.count(), 0)

    def test_device_fw_no_connection(self):
        try:
            self._create_device_firmware(device_connection=False)
        except ValidationError as e:
            self.assertIn('related connection', str(e))
        else:
            self.fail('ValidationError not raised')

    def test_invalid_board(self):
        image = FIRMWARE_IMAGE_MAP[self.TPLINK_4300_IMAGE]
        boards = image['boards']
        del image['boards']
        err = None
        try:
            self._create_firmware_image()
        except ValidationError as e:
            err = e
        image['boards'] = boards
        if err:
            self.assertIn('type', err.message_dict)
            self.assertIn('not find boards', str(err))
        else:
            self.fail('ValidationError not raised')

    def test_custom_image_type_present(self):
        t = FirmwareImage._meta.get_field('type')
        custom_images = app_settings.CUSTOM_OPENWRT_IMAGES
        self.assertEqual(t.choices[0][0], custom_images[0][0])

    def test_device_firmware_image_invalid_model(self):
        device_fw = self._create_device_firmware()
        different_img = self._create_firmware_image(
            build=device_fw.image.build, type=self.TPLINK_4300_IL_IMAGE
        )
        try:
            device_fw.image = different_img
            device_fw.full_clean()
        except ValidationError as e:
            self.assertIn('model do not match', str(e))
        else:
            self.fail('ValidationError not raised')

    def test_upgrade_operation_invalid_upgrade_options(self):
        device_fw = self._create_device_firmware()
        uo = UpgradeOperation(
            device=device_fw.device,
            image=device_fw.image,
        )
        with self.subTest('Test using invalid options'):
            uo.upgrade_options = {'invalid': True}
            with self.assertRaises(ValidationError) as error:
                uo.full_clean()
            self.assertEqual(
                error.exception.message_dict['__all__'],
                ['The upgrade options are invalid'],
            )

        with self.subTest('Test using mutually exclusive options'):
            uo.upgrade_options = {'c': True, 'n': True}
            with self.assertRaises(ValidationError) as error:
                uo.full_clean()
            self.assertEqual(
                error.exception.message_dict['upgrade_options'],
                ['The "-n" and "-c" options cannot be used together'],
            )

            uo.upgrade_options = {'o': True, 'n': True}
            with self.assertRaises(ValidationError) as error:
                uo.full_clean()
            self.assertEqual(
                error.exception.message_dict['upgrade_options'],
                ['The "-n" and "-o" options cannot be used together'],
            )

    def test_upgrade_operation_log_line(self):
        device_fw = self._create_device_firmware()
        uo = UpgradeOperation(device=device_fw.device, image=device_fw.image)
        uo.log_line('line1', save=False)
        uo.log_line('line2', save=False)
        self.assertEqual(uo.log, 'line1\nline2')
        try:
            uo.refresh_from_db()
        except UpgradeOperation.DoesNotExist:
            pass
        else:
            self.fail('item has been saved')

    def test_upgrade_operation_log_line_save(self):
        device_fw = self._create_device_firmware()
        uo = UpgradeOperation(device=device_fw.device, image=device_fw.image)
        uo.log_line('line1')
        uo.log_line('line2')
        uo.refresh_from_db()
        self.assertEqual(uo.log, 'line1\nline2')

    def test_permissions(self):
        admin = Group.objects.get(name='Administrator')
        operator = Group.objects.get(name='Operator')

        admin_permissions = [
            p['codename'] for p in admin.permissions.values('codename')
        ]
        operator_permissions = [
            p['codename'] for p in operator.permissions.values('codename')
        ]

        operators_read_only_admins_manage = [
            'build',
            'devicefirmware',
            'firmwareimage',
            'batchupgradeoperation',
            'upgradeoperation',
        ]
        admins_can_manage = ['category']
        manage_operations = ['add', 'change', 'delete']

        for action in manage_operations:
            for model_name in admins_can_manage:
                codename = '{}_{}'.format(action, model_name)
                self.assertIn(codename, admin_permissions)
                self.assertNotIn(codename, operator_permissions)

        for model_name in operators_read_only_admins_manage:
            codename = 'view_{}'.format(model_name)
            self.assertIn(codename, operator_permissions)

            for action in manage_operations:
                codename = '{}_{}'.format(action, model_name)
                self.assertIn(codename, admin_permissions)

    @capture_any_output()
    def test_create_for_device_validation_error(self):
        device_fw = self._create_device_firmware()
        device_fw.image.build.os = device_fw.device.os
        device_fw.image.build.save()
        result = DeviceFirmware.create_for_device(device_fw.device)
        self.assertIsNone(result)

    def test_upgrade_operation_retention_on_image_delete(self):
        device_fw = self._create_device_firmware()
        uo = UpgradeOperation.objects.create(
            device=device_fw.device, image=device_fw.image
        )
        FirmwareImage.objects.get(pk=device_fw.image.pk).delete()
        self.assertEqual(UpgradeOperation.objects.get(pk=uo.pk).image, None)


class TestModelsTransaction(TestUpgraderMixin, TransactionTestCase):
    _mock_updrade = 'openwisp_firmware_upgrader.upgraders.openwrt.OpenWrt.upgrade'
    _mock_connect = 'openwisp_controller.connection.models.DeviceConnection.connect'
    os = TestModels.os
    image_type = TestModels.image_type

    @mock.patch(_mock_updrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    def test_dry_run(self, *args):
        env = self._create_upgrade_env()
        # check pending upgrades
        result = BatchUpgradeOperation.dry_run(build=env['build1'])
        self.assertEqual(
            list(result['device_firmwares']),
            list(DeviceFirmware.objects.all().order_by('-created')),
        )
        self.assertEqual(list(result['devices']), [])
        # upgrade devices
        env['build1'].batch_upgrade(firmwareless=True)
        # check pending upgrades again
        result = BatchUpgradeOperation.dry_run(build=env['build1'])
        self.assertEqual(list(result['device_firmwares']), [])
        self.assertEqual(list(result['devices']), [])

    @mock.patch(_mock_updrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    def test_upgrade_related_devices(self, *args):
        env = self._create_upgrade_env()
        # check everything is as expected
        self.assertEqual(UpgradeOperation.objects.count(), 0)
        self.assertEqual(env['d1'].devicefirmware.image, env['image1a'])
        self.assertEqual(env['d2'].devicefirmware.image, env['image1b'])
        # upgrade all related
        env['build2'].batch_upgrade(firmwareless=False)
        # ensure image is changed
        env['d1'].devicefirmware.refresh_from_db()
        env['d2'].devicefirmware.refresh_from_db()
        self.assertEqual(env['d1'].devicefirmware.image, env['image2a'])
        self.assertEqual(env['d2'].devicefirmware.image, env['image2b'])
        # ensure upgrade operation objects have been created
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        batch_qs = BatchUpgradeOperation.objects.all()
        self.assertEqual(batch_qs.count(), 1)
        batch = batch_qs.first()
        self.assertEqual(batch.upgradeoperation_set.count(), 2)
        self.assertEqual(batch.build, env['build2'])
        self.assertEqual(batch.status, 'success')

    @mock.patch(_mock_updrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    def test_upgrade_firmwareless_devices(self, *args):
        env = self._create_upgrade_env(device_firmware=False)
        # check everything is as expected
        self.assertEqual(UpgradeOperation.objects.count(), 0)
        self.assertFalse(hasattr(env['d1'], 'devicefirmware'))
        self.assertFalse(hasattr(env['d2'], 'devicefirmware'))
        # upgrade all related
        env['build2'].batch_upgrade(firmwareless=True)
        env['d1'].refresh_from_db()
        env['d2'].refresh_from_db()
        self.assertEqual(env['d1'].devicefirmware.image, env['image2a'])
        self.assertEqual(env['d2'].devicefirmware.image, env['image2b'])
        # ensure upgrade operation objects have been created
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        batch_qs = BatchUpgradeOperation.objects.all()
        self.assertEqual(batch_qs.count(), 1)
        batch = batch_qs.first()
        self.assertEqual(batch.upgradeoperation_set.count(), 2)
        self.assertEqual(batch.build, env['build2'])
        self.assertEqual(batch.status, 'success')

    @mock.patch.object(upgrade_firmware, 'max_retries', 0)
    def test_batch_upgrade_failure(self):
        env = self._create_upgrade_env()
        with redirect_stdout(io.StringIO()):
            env['build2'].batch_upgrade(firmwareless=False)
        batch = BatchUpgradeOperation.objects.first()
        self.assertEqual(batch.status, 'failed')
        self.assertEqual(BatchUpgradeOperation.objects.count(), 1)

    @mock.patch(_mock_updrade, return_value=True)
    @mock.patch(_mock_connect, return_value=True)
    def test_upgrade_related_devices_existing_fw(self, *args):
        env = self._create_upgrade_env()
        self.assertEqual(UpgradeOperation.objects.count(), 0)
        self.assertEqual(env['d1'].devicefirmware.image, env['image1a'])
        self.assertEqual(env['d2'].devicefirmware.image, env['image1b'])
        env['d1'].devicefirmware.installed = False
        env['d1'].devicefirmware.save(upgrade=False)
        env['d2'].devicefirmware.installed = False
        env['d2'].devicefirmware.save(upgrade=False)
        env['build1'].batch_upgrade(firmwareless=False)
        env['d1'].devicefirmware.refresh_from_db()
        env['d2'].devicefirmware.refresh_from_db()
        self.assertEqual(env['d1'].devicefirmware.image, env['image1a'])
        self.assertEqual(env['d2'].devicefirmware.image, env['image1b'])
        self.assertEqual(UpgradeOperation.objects.count(), 2)
        batch_qs = BatchUpgradeOperation.objects.all()
        self.assertEqual(batch_qs.count(), 1)
        batch = batch_qs.first()
        self.assertEqual(batch.upgradeoperation_set.count(), 2)
        self.assertEqual(batch.build, env['build1'])
        self.assertEqual(batch.status, 'success')

    def test_upgrade_retried(self):
        env = self._create_upgrade_env()
        try:
            with redirect_stdout(io.StringIO()):
                env['build2'].batch_upgrade(firmwareless=False)
        except Retry:
            pass
        except Exception as e:
            self.fail(f'Expected retry, got {e.__class__} instead')
        else:
            self.fail('Retry exception not raised')
        self.assertEqual(BatchUpgradeOperation.objects.count(), 1)
        batch = BatchUpgradeOperation.objects.first()
        self.assertEqual(batch.status, 'in-progress')

    def test_device_fw_not_created_on_device_connection_save(self):
        org = self._get_org()
        category = self._get_category(organization=org)
        build1 = self._create_build(category=category, version='0.1', os=self.os)
        image1a = self._create_firmware_image(build=build1, type=self.image_type)

        with self.subTest("Device doesn't define os"):
            d1 = self._create_device_with_connection(os='', model=image1a.boards[0])
            self.assertEqual(DeviceConnection.objects.count(), 1)
            self.assertEqual(Device.objects.count(), 1)
            self.assertEqual(DeviceFirmware.objects.count(), 0)
            d1.delete(check_deactivated=False)
            Credentials.objects.all().delete()

        with self.subTest("Device doesn't define model"):
            d1 = self._create_device_with_connection(os=self.os, model='')
            self.assertEqual(DeviceConnection.objects.count(), 1)
            self.assertEqual(Device.objects.count(), 1)
            self.assertEqual(DeviceFirmware.objects.count(), 0)
            d1.delete(check_deactivated=False)
            Credentials.objects.all().delete()

        build1.os = None
        build1.save()

        with self.subTest("Build doesn't define os"):
            d1 = self._create_device_with_connection(
                model=image1a.boards[0], os=self.os
            )
            self.assertEqual(DeviceConnection.objects.count(), 1)
            self.assertEqual(Device.objects.count(), 1)
            self.assertEqual(DeviceFirmware.objects.count(), 0)

    def test_device_fw_created_on_device_connection_save(self):
        self.assertEqual(DeviceFirmware.objects.count(), 0)
        self.assertEqual(Device.objects.count(), 0)
        org = self._get_org()
        category = self._get_category(organization=org)
        build1 = self._create_build(category=category, version='0.1', os=self.os)
        image1a = self._create_firmware_image(build=build1, type=self.image_type)
        self._create_device_with_connection(os=self.os, model=image1a.boards[0])
        self.assertEqual(Device.objects.count(), 1)
        self.assertEqual(DeviceFirmware.objects.count(), 1)
        self.assertEqual(DeviceConnection.objects.count(), 1)

    def test_delete_firmware_image_file(self):
        file_storage_backend = FirmwareImage.file.field.storage

        with self.subTest('Test deleting object deletes file'):
            image = self._create_firmware_image()
            file_name = image.file.name
            image.delete()
            self.assertEqual(file_storage_backend.exists(file_name), False)

        with self.subTest('Test deleting object with a deleted file'):
            image = self._create_firmware_image()
            file_name = image.file.name
            # Delete the file from the storage backend before
            # deleting the object
            file_storage_backend.delete(file_name)
            self.assertNotEqual(image.file, None)
            image.delete()



from django.utils.translation import gettext_lazy as _

from openwisp_users.multitenancy import (
    MultitenantOrgFilter,
    MultitenantRelatedOrgFilter,
)

from .swapper import load_model

Build = load_model('Build')
Category = load_model('Category')


class CategoryFilter(MultitenantRelatedOrgFilter):
    field_name = 'category'
    parameter_name = 'category_id'
    title = _('category')


class CategoryOrganizationFilter(MultitenantOrgFilter):
    parameter_name = 'category__organization'
    rel_model = Category


class BuildCategoryFilter(MultitenantRelatedOrgFilter):
    field_name = 'category'
    parameter_name = 'build__category'
    title = _('category')
    rel_model = Build


class BuildCategoryOrganizationFilter(MultitenantOrgFilter):
    parameter_name = 'build__category__organization'
    rel_model = Category



# Generated by Django 3.0.8 on 2020-07-07 17:00

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('firmware_upgrader', '0004_batch_upgrade_operation_idle_status'),
    ]

    operations = [
        migrations.AddField(
            model_name='build',
            name='os',
            field=models.CharField(
                blank=True,
                help_text='OS identifier as presented by the device, used to automatically recognize the firmware image used by new devices that register into the system',
                max_length=64,
                null=True,
                verbose_name='OS identifier',
            ),
        ),
    ]



import os
from unittest import mock

import swapper
from django.conf import settings
from django.core.files.uploadedfile import SimpleUploadedFile

from openwisp_controller.connection.tests.utils import CreateConnectionsMixin

from ..swapper import load_model

Build = load_model('Build')
Category = load_model('Category')
FirmwareImage = load_model('FirmwareImage')
DeviceFirmware = load_model('DeviceFirmware')
DeviceFirmware = load_model('DeviceFirmware')
Credentials = swapper.load_model('connection', 'Credentials')


class TestUpgraderMixin(CreateConnectionsMixin):
    FAKE_IMAGE_PATH = os.path.join(settings.PRIVATE_STORAGE_ROOT, 'fake-img.bin')
    FAKE_IMAGE_PATH2 = os.path.join(settings.PRIVATE_STORAGE_ROOT, 'fake-img2.bin')
    TPLINK_4300_IMAGE = 'ath79-generic-tplink_tl-wdr4300-v1-squashfs-sysupgrade.bin'
    TPLINK_4300_IL_IMAGE = (
        'ath79-generic-tplink_tl-wdr4300-v1-il-squashfs-sysupgrade.bin'
    )

    def tearDown(self):
        for fw in FirmwareImage.objects.all():
            fw.delete()

    def _get_build(self, version="0.1", **kwargs):
        opts = {"version": version}
        opts.update(kwargs)
        try:
            return Build.objects.get(**opts)
        except Build.DoesNotExist:
            return self._create_build(**opts)

    def _get_category(self, cat_name="Test Category", **kwargs):
        opts = {"name": cat_name}
        opts.update(kwargs)
        try:
            return Category.objects.get(**opts)
        except Category.DoesNotExist:
            return self._create_category(**opts)

    def _create_category(self, **kwargs):
        opts = dict(name='Test Category')
        opts.update(kwargs)
        if 'organization' not in opts:
            opts['organization'] = self._get_org()
        c = Category(**opts)
        c.full_clean()
        c.save()
        return c

    def _create_build(self, **kwargs):
        opts = dict(version='0.1')
        opts.update(kwargs)
        category_opts = {}
        if 'organization' in opts:
            category_opts = {'organization': opts.pop('organization')}
        if 'category' not in opts:
            opts['category'] = self._get_category(**category_opts)
        b = Build(**opts)
        b.full_clean()
        b.save()
        return b

    def _create_firmware_image(self, **kwargs):
        opts = dict(type=self.TPLINK_4300_IMAGE)
        opts.update(kwargs)
        build_opts = {}
        if 'organization' in opts:
            build_opts['organization'] = opts.pop('organization')
        if 'build' not in opts:
            opts['build'] = self._get_build(**build_opts)
        if 'file' not in opts:
            opts['file'] = self._get_simpleuploadedfile()
        fw = FirmwareImage(**opts)
        fw.full_clean()
        fw.save()
        return fw

    def _get_simpleuploadedfile(self, filename=None):
        if not filename:
            filename = self.FAKE_IMAGE_PATH
        with open(filename, 'rb') as f:
            image = f.read()
        return SimpleUploadedFile(
            name=f'openwrt-{self.TPLINK_4300_IMAGE}',
            content=image,
            content_type='application/octet-stream',
        )

    def _create_device_firmware(self, upgrade=False, device_connection=True, **kwargs):
        opts = dict()
        opts.update(kwargs)
        if 'image' not in opts:
            opts['image'] = self._create_firmware_image()
        if 'device' not in opts:
            org = opts['image'].build.category.organization
            opts['device'] = self._create_device(organization=org)
            self._create_config(device=opts['device'])
        if device_connection:
            self._create_device_connection(device=opts['device'])
        device_fw = DeviceFirmware(**opts)
        device_fw.full_clean()
        device_fw.save(upgrade=upgrade)
        return device_fw

    def _create_upgrade_env(
        self, device_firmware=True, upgrade_operation=False, **kwargs
    ):
        org = kwargs.pop('organization', self._get_org())
        category = kwargs.pop('category', self._get_category(organization=org))
        build1 = self._create_build(category=category, version='0.1')
        image1a = self._create_firmware_image(build=build1, type=self.TPLINK_4300_IMAGE)
        image1b = self._create_firmware_image(
            build=build1, type=self.TPLINK_4300_IL_IMAGE
        )
        # create devices
        d1 = self._create_device(
            name='device1',
            organization=org,
            mac_address='00:22:bb:33:cc:44',
            model=image1a.boards[0],
        )
        d2 = self._create_device(
            name='device2',
            organization=org,
            mac_address='00:11:bb:22:cc:33',
            model=image1b.boards[0],
        )
        ssh_credentials = self._get_credentials(organization=None)
        self._create_config(device=d1)
        self._create_config(device=d2)
        self._create_device_connection(device=d1, credentials=ssh_credentials)
        self._create_device_connection(device=d2, credentials=ssh_credentials)

        # create a new firmware build
        build2 = self._create_build(category=category, version='0.2')
        image2a = self._create_firmware_image(build=build2, type=self.TPLINK_4300_IMAGE)
        image2b = self._create_firmware_image(
            build=build2, type=self.TPLINK_4300_IL_IMAGE
        )
        data = {
            'build1': build1,
            'build2': build2,
            'd1': d1,
            'd2': d2,
            'image1a': image1a,
            'image1b': image1b,
            'image2a': image2a,
            'image2b': image2b,
        }
        # force create device firmware (optional)
        if device_firmware:
            device_fw1 = self._create_device_firmware(
                device=d1,
                image=image1a,
                upgrade=upgrade_operation,
                device_connection=False,
            )
            device_fw2 = self._create_device_firmware(
                device=d2,
                image=image1b,
                upgrade=upgrade_operation,
                device_connection=False,
            )
            data.update(
                {
                    'device_fw1': device_fw1,
                    'device_fw2': device_fw2,
                }
            )
        return data

    def _create_firmwareless_device(self, organization):
        d = self._create_device(
            name='firmwareless',
            mac_address='01:12:23:44:55:66',
            organization=organization,
        )
        self._create_config(device=d)
        self._create_device_connection(
            device=d, credentials=Credentials.objects.first()
        )
        return d

    def _create_device_with_connection(self, **kwargs):
        d1 = self._create_device(**kwargs)
        self._create_config(device=d1)
        self._create_device_connection(device=d1)
        return d1


def spy_mock(method, pre_action):
    magicmock = mock.MagicMock()

    def wrapper(*args, **kwargs):
        magicmock(*args, **kwargs)
        pre_action(*args, **kwargs)
        return method(*args, **kwargs)

    wrapper.mock = magicmock
    return wrapper



import os
import re
from hashlib import sha256
from time import sleep

import jsonschema
from billiard import Process, Queue
from django.utils.translation import gettext_lazy as _

from openwisp_controller.connection.exceptions import NoWorkingDeviceConnectionError

from ..exceptions import (
    FirmwareUpgradeOptionsException,
    ReconnectionFailed,
    RecoverableFailure,
    UpgradeAborted,
    UpgradeNotNeeded,
)
from ..settings import OPENWRT_SETTINGS


class OpenWrt(object):
    CHECKSUM_FILE = '/etc/openwisp/firmware_checksum'
    REMOTE_UPLOAD_DIR = '/tmp'
    RECONNECT_DELAY = OPENWRT_SETTINGS.get('reconnect_delay', 180)
    RECONNECT_RETRY_DELAY = OPENWRT_SETTINGS.get('reconnect_retry_delay', 20)
    RECONNECT_MAX_RETRIES = OPENWRT_SETTINGS.get('reconnect_max_retries', 35)
    UPGRADE_TIMEOUT = OPENWRT_SETTINGS.get('upgrade_timeout', 90)
    UPGRADE_COMMAND = '{sysupgrade} -v {flags} {path}'
    # path to sysupgrade command
    _SYSUPGRADE = '/sbin/sysupgrade'
    SCHEMA = {
        'type': 'object',
        'properties': {
            'c': {
                'type': 'boolean',
                'title': _('Attempt to preserve all changed files in /etc/ (-c)'),
                'format': 'checkbox',
                'default': True,
            },
            'o': {
                'type': 'boolean',
                'title': _(
                    'Attempt to preserve all changed files in /, except those from '
                    'packages but including changed confs. (-o)'
                ),
                'format': 'checkbox',
            },
            'n': {
                'type': 'boolean',
                'title': _('Do not save configuration over reflash (-n)'),
                'format': 'checkbox',
            },
            'u': {
                'type': 'boolean',
                'title': _(
                    'Skip from backup files that are equal to those in /rom (-u)'
                ),
                'format': 'checkbox',
            },
            'p': {
                'type': 'boolean',
                'title': _(
                    'Do not attempt to restore the partition table after flash. (-p)'
                ),
                'format': 'checkbox',
            },
            'k': {
                'type': 'boolean',
                'title': _(
                    'Include in backup a list of current installed packages at'
                    ' /etc/backup/installed_packages.txt (-k)'
                ),
                'format': 'checkbox',
            },
            'F': {
                'type': 'boolean',
                'title': _(
                    'Flash image even if image checks fail, this is dangerous! (-F)'
                ),
                'format': 'checkbox',
            },
        },
        'additionalProperties': False,
    }

    log_lines = None

    _false_positives = [
        'Command failed: ubus call system sysupgrade '
        '\{ "prefix": "\\\/tmp\\\/root", "path": "[^"]*", '
        '"backup": "\\\/tmp\\\/sysupgrade.tgz", '
        '"command": "\\\/lib\\\/upgrade\\\/do_stage2", '
        '"options": \{ "save_partitions": 1 \} \}',
        'Command failed: ubus call system sysupgrade '
        '\{ "prefix": "\\\/tmp\\\/root", "path": "[^"]*", '
        '"command": "\\\/lib\\\/upgrade\\\/do_stage2", '
        '"options": \{ "save_partitions": 1 \} \}',
    ]

    def __init__(self, upgrade_operation, connection):
        self.upgrade_operation = upgrade_operation
        self.connection = connection
        self._non_critical_services_stopped = False

    @classmethod
    def validate_upgrade_options(cls, upgrade_options):
        jsonschema.Draft4Validator(cls.SCHEMA).validate(upgrade_options)
        if upgrade_options.get('n', False):
            if upgrade_options.get('o', False):
                raise FirmwareUpgradeOptionsException(
                    {
                        'upgrade_options': _(
                            'The "-n" and "-o" options cannot be used together'
                        )
                    }
                )
            if upgrade_options.get('c', False):
                raise FirmwareUpgradeOptionsException(
                    {
                        'upgrade_options': _(
                            'The "-n" and "-c" options cannot be used together'
                        )
                    }
                )

    def log(self, value, save=True):
        self.upgrade_operation.log_line(value, save=save)

    def connect(self):
        """
        Wrapper method to call "connect" method of the connection object
        """
        return self.connection.connect()

    def disconnect(self):
        """
        Wrapper method to call "disconnect" method of the connection object
        """
        return self.connection.disconnect()

    def exec_command(self, *args, **kwargs):
        """
        Wrapper method to call "exec_command" method of the connection object
        """
        return self.connection.connector_instance.exec_command(*args, **kwargs)

    def upload(self, image_file, remote_path):
        """
        Wrapper method to call "upload" method of the connection object
        """
        self.check_memory(image_file)
        try:
            self.connection.connector_instance.upload(image_file, remote_path)
        except Exception as e:
            raise RecoverableFailure(str(e))

    def upgrade(self, image):
        self._test_connection()
        checksum = self._test_checksum(image)
        remote_path = self.get_remote_path(image)
        self.upload(image.file, remote_path)
        self._test_image(remote_path)
        self._reflash(remote_path)
        self._write_checksum(checksum)

    def _test_connection(self):
        result = self.connect()
        if not result:
            raise RecoverableFailure('Connection failed')
        self.log(_('Connection successful, starting upgrade...'))

    _non_critical_services = [
        'uhttpd',
        'dnsmasq',
        'openwisp_config',
        'openwisp-config',
        'openwisp-monitoring',
        'cron',
        'rpcd',
        'rssileds',
        'odhcpd',
        'log',
    ]

    def check_memory(self, image_file):
        """
        Tries to free up memory before upgrading
        """
        self._free_memory()
        # if there's enouogh available memory, proceed
        current_free_memory = self._get_free_memory()
        if image_file.size < current_free_memory:
            return
        file_size_mib = self._get_mib(image_file.size)
        free_memory_mib = self._get_mib(current_free_memory)
        # otherwise try to free up some more memory by stopping services
        self.log(
            _(
                'The image size ({file_size_mib} MiB) is greater '
                'than the available memory on the system ({free_memory_mib} MiB).\n'
                'For this reason the upgrade procedure will try to free up '
                'memory by stopping non critical services.\n'
                'WARNING: it is recommended to reboot the device is the upgrade '
                'fails unexpectedly because these services will not be restarted '
                'automatically.\n'
                'NOTE: The reboot can be avoided if the status of the upgrade becomes '
                '"aborted" because in this case the system will restart the '
                'services automatically.'.format(
                    file_size_mib=file_size_mib, free_memory_mib=free_memory_mib
                )
            )
        )
        self._stop_non_critical_services()
        self._free_memory()
        # check memory again
        # this time abort if there's still not enough free memory
        current_free_memory = self._get_free_memory()
        free_memory_mib = self._get_mib(current_free_memory)
        if image_file.size < current_free_memory:
            self.log(
                _(
                    'Enough available memory was freed up on the system '
                    '({0} MiB)!\n'
                    'Proceeding to upload of the image file...'.format(free_memory_mib)
                )
            )
        else:
            self.log(
                _(
                    'There is still not enough available memory on '
                    'the system ({0} MiB).\n'
                    'Starting non critical services again...'.format(free_memory_mib)
                )
            )
            self._start_non_critical_services()
            self.log(_('Non critical services started, aborting upgrade.'))
            raise UpgradeAborted()

    def _get_mib(self, value):
        """
        Converts bytes to megabytes
        """
        if value == 0:
            return value
        _MiB = 1048576
        return round(value / _MiB, 2)

    def _get_free_memory(self):
        """
        Tries to get the available memory
        If that fails it falls back to use MemFree (should happen only on older systems)
        """
        meminfo_grep = 'cat /proc/meminfo | grep'
        output, exit_code = self.exec_command(
            f'{meminfo_grep} MemAvailable', exit_codes=[0, 1]
        )
        if exit_code == 1:
            output, exit_code = self.exec_command(f'{meminfo_grep} MemFree')
        parts = output.split()
        return int(parts[1]) * 1024

    def _free_memory(self):
        """
        Attempts to free up some memory without stopping any service.
        """
        # remove OPKG index
        self.exec_command('rm -rf /tmp/opkg-lists/')
        # free internal cache
        self.exec_command('sync && echo 3 > /proc/sys/vm/drop_caches')

    def _stop_non_critical_services(self):
        """
        Stops non critical services in order to free up memory.
        """
        for service in self._non_critical_services:
            initd = f'/etc/init.d/{service}'
            self.exec_command(
                f'test -f {initd} && {initd} stop', raise_unexpected_exit=False
            )
        self.exec_command('test -f /sbin/wifi && /sbin/wifi down')
        self._non_critical_services_stopped = True

    def _start_non_critical_services(self):
        """
        Starts again non critical services.
        To be used if an upgrade operation is aborted.
        """
        for service in self._non_critical_services:
            initd = f'/etc/init.d/{service}'
            self.exec_command(
                f'test -f {initd} && {initd} start', raise_unexpected_exit=False
            )
        self.exec_command('test -f /sbin/wifi && /sbin/wifi up')
        self._non_critical_services_stopped = False

    def get_remote_path(self, image):
        # discard directory info from image name
        filename = image.name.split('/')[-1]
        return os.path.join(self.REMOTE_UPLOAD_DIR, filename)

    def get_upgrade_command(self, path):
        # Build flags for the command
        flags = []
        for key, value in self.upgrade_operation.upgrade_options.items():
            if value is True:
                flags.append(f'-{key}')
        # Enforce defaults
        for key, property in self.SCHEMA['properties'].items():
            if self.upgrade_operation.upgrade_options.get(
                key, None
            ) is None and property.get('default'):
                flags.append(f'-{key}')
        flags = ' '.join(flags)
        return self.UPGRADE_COMMAND.format(
            sysupgrade=self._SYSUPGRADE, flags=flags, path=path
        )

    def _test_checksum(self, image):
        """
        prevents the upgrade if an identical checksum signature file is found on
        the device, which indicates the upgrade has already been performed previously
        """
        # calculate firmware image checksum
        checksum = sha256(image.read()).hexdigest()
        image.seek(0)
        # test for presence of firmware checksum signature file
        output, exit_code = self.exec_command(
            f'test -f {self.CHECKSUM_FILE}', exit_codes=[0, 1]
        )
        if exit_code == 0:
            self.log(_('Image checksum file found'), save=False)
            cat = f'cat {self.CHECKSUM_FILE}'
            output, code = self.exec_command(cat)
            if checksum == output.strip():
                message = _(
                    'Firmware already upgraded previously. '
                    'Identical checksum found in the filesystem, '
                    'upgrade not needed.'
                )
                self.log(message)
                self.disconnect()
                raise UpgradeNotNeeded(message)
            else:
                self.log(
                    _(
                        'Checksum different, proceeding with '
                        'the upload of the new image...'
                    )
                )
        else:
            self.log(
                _(
                    'Image checksum file not found, proceeding '
                    'with the upload of the new image...'
                )
            )
        return checksum

    def _test_image(self, path):
        try:
            self.exec_command(f'{self._SYSUPGRADE} --test {path}')
        except Exception as e:
            self.log(str(e), save=False)
            # if non critical services were stopped to free up memory, restart them
            if self._non_critical_services_stopped:
                self.log(_('Starting non critical services again...'))
                self._start_non_critical_services()
            self.disconnect()
            raise UpgradeAborted()
        self.log(
            _(
                'Sysupgrade test passed successfully, '
                'proceeding with the upgrade operation...'
            )
        )

    def _reflash(self, path):
        """
        this method will execute the reflashing operation in another process
        because the SSH connection may hang indefinitely while reflashing
        and would block the program; setting a timeout to `exec_command`
        doesn't seem to take effect on some OpenWrt versions
        so at least we can stop the process using
        `subprocess.join(timeout=self.UPGRADE_TIMEOUT)`
        """
        self.disconnect()
        self.log(_('Upgrade operation in progress...'))

        failure_queue = Queue()
        subprocess = Process(
            target=self._call_reflash_command,
            args=[self, path, self.UPGRADE_TIMEOUT, failure_queue],
        )
        subprocess.start()
        subprocess.join(timeout=self.UPGRADE_TIMEOUT)

        # if the child process catched an exception, raise it here in the
        # parent so it will be logged and will flag the upgrade as failed
        if not failure_queue.empty():
            raise failure_queue.get()
        failure_queue.close()

        self.upgrade_operation.refresh_from_db()
        self.log(
            _(
                'SSH connection closed, will wait {0} '
                'seconds before attempting to reconnect...'.format(self.RECONNECT_DELAY)
            )
        )
        sleep(self.RECONNECT_DELAY)
        # kill the subprocess if it has hanged
        if subprocess.is_alive():
            subprocess.terminate()
            subprocess.join()

    @classmethod
    def _call_reflash_command(cls, upgrader, path, timeout, failure_queue):
        try:
            upgrader.connect()
            command = upgrader.get_upgrade_command(path)
            # remove persistent checksum if present (introduced in openwisp-config 0.6.0)
            # otherwise the device will not download the configuration again after reflash
            upgrader.exec_command(
                'rm /etc/openwisp/checksum 2> /dev/null', exit_codes=[0, -1, 1]
            )
            output, exit_code = upgrader.exec_command(
                command, timeout=timeout, exit_codes=[0, -1]
            )
            upgrader.log(output)
        except Exception as e:
            # In some cases, for some unknown reason, the sysupgrade command
            # returns a non zero exit code, but it is carried out anyway.
            # This is a workaround to recognize this false positive and ignore it.
            error = str(e)
            for pattern in cls._false_positives:
                if re.search(pattern, error):
                    break
            else:
                failure_queue.put(e)
        upgrader.disconnect()

    def _refresh_addresses(self):
        """
        reloads the device info from the DB to
        handle cases in which the IP has changed
        """
        self.connection.device.refresh_from_db()
        self.connection = self.connection.get_working_connection(self.connection.device)
        self.addresses = self.connection.get_addresses()

    def _log_reconnecting_error(self, attempt):
        addresses = ', '.join(self.addresses)
        self.log(
            _(
                'Trying to reconnect to device at {addresses} (attempt n.{attempt})...'.format(
                    addresses=addresses, attempt=attempt
                )
            ),
            save=False,
        )

    def _write_checksum(self, checksum):
        for attempt in range(1, self.RECONNECT_MAX_RETRIES + 1):
            try:
                self._refresh_addresses()
            except NoWorkingDeviceConnectionError as error:
                if error.connection:
                    self.addresses = error.connection.get_addresses()
                self._log_reconnecting_error(attempt)
                if not str(error):
                    error = _('connection failed')
                self.log(
                    _(
                        'Device not reachable yet, ({0}).\n'
                        'retrying in {1} seconds...'.format(
                            error, self.RECONNECT_RETRY_DELAY
                        )
                    )
                )
                sleep(self.RECONNECT_RETRY_DELAY)
                continue
            self._log_reconnecting_error(attempt)
            self.log(_('Connected! Writing checksum ' f'file to {self.CHECKSUM_FILE}'))
            checksum_dir = os.path.dirname(self.CHECKSUM_FILE)
            self.exec_command(f'mkdir -p {checksum_dir}')
            self.exec_command(f'echo {checksum} > {self.CHECKSUM_FILE}')
            self.disconnect()
            self.log(_('Upgrade completed successfully.'))
            return
        # if all previous attempts failed
        raise ReconnectionFailed(
            'Giving up, device not reachable anymore after upgrade'
        )



#!/usr/bin/env python
from setuptools import find_packages, setup

from openwisp_firmware_upgrader import get_version

setup(
    name='openwisp-firmware-upgrader',
    version=get_version(),
    license='GPL3',
    author='OpenWISP',
    author_email='support@openwisp.io',
    description='Firmware upgrader module of OpenWISP',
    long_description=open('README.rst').read(),
    url='http://openwisp.org',
    download_url='https://github.com/openwisp/openwisp-firmware-upgrader/releases',
    platforms=['Platform Independent'],
    keywords=['django', 'netjson', 'networking', 'openwisp', 'firmware'],
    packages=find_packages(exclude=['tests*', 'docs*']),
    include_package_data=True,
    zip_safe=False,
    install_requires=[
        'openwisp-controller~=1.1.0',
        'django-private-storage~=3.1.0',
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable ',
        'Environment :: Web Environment',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: System :: Networking',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Framework :: Django',
        'Programming Language :: Python :: 3',
    ],
)
